"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
A number of functions around Revit independent tags elbow properties.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed. 
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits; 
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#


from Autodesk.Revit.DB import LeaderEndCondition

from duHast.Revit.Common.revit_version import get_revit_version_number
from duHast.Revit.Common.Geometry.points import get_point_as_doubles

ELBOW_LOCATION = "elbow_location"
LEADER_END = "leader_end"
LEADER_REFERENCE = "leader_reference"
LEADER_ELEMENT_REFERENCE_ID = "leader_element_reference_id"
LEADER_LINKED_ELEMENT_REFERENCE_ID = "leader_linked_element_reference_id"


def get_elbow_properties(doc, tag, points_as_double=False):
    """
    Returns elbow properties depending on revit version.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param tag: The tag element
    :type tag: Autodesk.Revit.DB.IndependentTag

    :return: A list of dictionaries with 3 keys: ELBOW_LOCATION, LEADER_END and LEADER_REFERENCE.
    :rtype: [{str:Autodesk.Revit.DB.XYZ, str:Autodesk.Revit.DB.XYZ, str:Autodesk.Revit.DB Reference }]
    """

    # get the revit version:
    revit_version = get_revit_version_number(doc)
    data = None
    if revit_version <= 2021:
        data = get_elbow_properties_2021(tag, points_as_double)
    elif revit_version >= 2022:
        data = get_elbow_properties_2022(tag, points_as_double)
    return data


def get_elbow_properties_2021(tag, points_as_double):
    """
    Get elbow properties in revit versions up to revit 2022

    :param tag: The tag element
    :type tag: Autodesk.Revit.DB.IndependentTag

    :return: A list of a single dictionaries with 3 keys: ELBOW_LOCATION, LEADER_END and LEADER_REFERENCE.
    :rtype: [{str:Autodesk.Revit.DB.XYZ, str:Autodesk.Revit.DB.XYZ, str:Autodesk.Revit.DB Reference }]
    """

    data = []
    elbow_properties = {}
    try:
        if tag.HasElbow:
            elbow_properties[LEADER_REFERENCE] = {
                LEADER_ELEMENT_REFERENCE_ID: None,
                LEADER_LINKED_ELEMENT_REFERENCE_ID: None,
            }
            if points_as_double:
                elbow_properties[ELBOW_LOCATION] = get_point_as_doubles(tag.LeaderElbow)
            else:
                elbow_properties[ELBOW_LOCATION] = tag.LeaderElbow
            if tag.LeaderEndCondition == LeaderEndCondition.Free:
                if points_as_double:
                    elbow_properties[LEADER_END] = get_point_as_doubles(tag.LeaderEnd)
                else:
                    elbow_properties[LEADER_END] = tag.LeaderEnd
            else:
                elbow_properties[LEADER_END] = None
        else:
            elbow_properties[ELBOW_LOCATION] = None
            elbow_properties[LEADER_END] = None
    except:
        elbow_properties[ELBOW_LOCATION] = None
        elbow_properties[LEADER_END] = None

    data.append(elbow_properties)
    return data


def get_elbow_properties_2022(tag, points_as_double):
    """
    Get elbow properties in revit versions from revit 2023 onwards

    :param tag: The tag element
    :type tag: Autodesk.Revit.DB.IndependentTag

    :return: A list of dictionaries with 3 keys: ELBOW_LOCATION, LEADER_END and LEADER_REFERENCE.
    :rtype: [{str:Autodesk.Revit.DB.XYZ, str:Autodesk.Revit.DB.XYZ, str:Autodesk.Revit.DB Reference }]
    """

    data = []
    tagged_references = tag.GetTaggedReferences()
    for tag_ref in tagged_references:
        elbow_properties = {}
        # store references to linked element or the element in same model which got tagged only
        elbow_properties[LEADER_REFERENCE] = {
            LEADER_ELEMENT_REFERENCE_ID: tag_ref.ElementId.IntegerValue,
            LEADER_LINKED_ELEMENT_REFERENCE_ID: tag_ref.LinkedElementId.IntegerValue,
        }
        try:
            if tag.HasLeaderElbow(tag_ref):
                if points_as_double:
                    elbow_properties[ELBOW_LOCATION] = get_point_as_doubles(
                        tag.GetLeaderElbow(tag_ref)
                    )
                else:
                    elbow_properties[ELBOW_LOCATION] = tag.GetLeaderElbow(tag_ref)
                if tag.LeaderEndCondition == LeaderEndCondition.Free:
                    if points_as_double:
                        elbow_properties[LEADER_END] = get_point_as_doubles(
                            tag.GetLeaderEnd(tag_ref)
                        )
                    else:
                        elbow_properties[LEADER_END] = tag.GetLeaderEnd(tag_ref)
                else:
                    elbow_properties[LEADER_END] = None
            else:
                elbow_properties[ELBOW_LOCATION] = None
                elbow_properties[LEADER_END] = None
        except:
            elbow_properties[ELBOW_LOCATION] = None
            elbow_properties[LEADER_END] = None
        data.append(elbow_properties)
    return data



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Data category and filter override instances to storage object helper functions.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

from duHast.Revit.Views.Objects.category_override_storage import RevitCategoryOverride
from duHast.Revit.Views.Objects.filter_override_storage import RevitFilterOverride
from duHast.Revit.Categories.categories_model import get_category_by_names
from duHast.Revit.Views.filters import get_filter_by_name
from duHast.Revit.Views.Utility.convert_data_to_revit_override import (
    convert_to_revit_graphic_override,
)


def convert_to_category_override_storage(doc, category_data_instance):
    """
    Retrieves the corresponding category object from the Revit document based on the category names provided in the category data instance.

    Args:
        doc (Revit Document): The Revit document object.
        category_data_instance (CategoryData): An instance of the CategoryData class representing the category and subcategory names, and visibility information.

    Returns:
        RevitCategoryOverride or None: The created RevitCategoryOverride object if a match is found, otherwise None.
    """
    return_value = None
    # get the category
    category_in_model = get_category_by_names(
        doc=doc,
        main_category_name=category_data_instance.main_category_name,
        sub_category_name=category_data_instance.sub_category_name,
    )
    # get the revit override
    revit_override = convert_to_revit_graphic_override(
        doc=doc, data_override=category_data_instance, is_filter_override=False
    )

    if category_in_model:
        return_value = RevitCategoryOverride(
            main_category_name=category_data_instance.main_category_name,
            sub_category_name=category_data_instance.sub_category_name,
            category=category_in_model,
            category_id=category_in_model.Id,
            revit_override=revit_override,
            is_category_hidden=not category_data_instance.is_visible, #need to be inverted :(((
        )
    return return_value


def convert_to_category_override_storage_objects(doc, category_data_objects):
    """
    Converts a list of category data objects into category override storage objects.

    Args:
        doc (Revit Document): The Revit document object.
        category_data_objects (list): A list of category data objects.

    Returns:
        list: A list of category override storage objects representing the converted category data objects.
    """
    return_value = []
    for cat_instance in category_data_objects:
        converted = convert_to_category_override_storage(
            doc=doc, category_data_instance=cat_instance
        )
        if converted:
            return_value.append(converted)
    return return_value


def convert_to_filter_override_storage(doc, filter_data_instance):
    """
    Converts a filter data instance into a `RevitFilterOverride` object.

    Args:
        doc (Revit Document): The current Revit model document.
        filter_data_instance (FilterDataInstance): The filter data instance to be converted.

    Returns:
        RevitFilterOverride: The converted filter override object, or None if the filter does not exist in the model.
    """
    return_value = None
    # get the filter
    filter_in_model = get_filter_by_name(
        doc=doc, filter_name=filter_data_instance.filter_name
    )

    # get the revit override
    revit_override = convert_to_revit_graphic_override(
        doc=doc, data_override=filter_data_instance, is_filter_override=True
    )

    if filter_in_model:
        return_value = RevitFilterOverride(
            filter_name=filter_data_instance.filter_name,
            filter_id=filter_in_model.Id,
            filter=filter_in_model,
            revit_override=revit_override,
            is_filter_visible=filter_data_instance.is_visible,
            is_filter_enabled=filter_data_instance.is_enabled,
        )

    return return_value


def convert_to_filter_override_storage_objects(doc, filter_data_objects):
    """
    Converts a list of filter data objects into a list of RevitFilterOverride objects.

    Args:
        doc (Revit Document): The current Revit model document.
        filter_data_objects (list): A list of filter data objects to be converted.

    Returns:
        list: A list of RevitFilterOverride objects that represent the converted filter data objects.
    """
    return_value = []
    for filter_instance in filter_data_objects:
        converted = convert_to_filter_override_storage(
            doc=doc, filter_data_instance=filter_instance
        )
        if converted:
            return_value.append(converted)
    return return_value



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Revit solids helper functions
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed. 
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits; 
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#


from Autodesk.Revit.DB import BoundingBoxXYZ, ElementId, Options, Solid
from duHast.Data.Objects.Properties.Geometry.from_revit_conversion import (
    convert_solid_to_flattened_2d_points,
)

from duHast.Revit.Common.Geometry.geometry import merge_bounding_box_xyz


def get_2d_points_from_solid(element):
    """
    Returns a list of lists of data geometry instances representing the flattened (2D geometry) of the Element
    List of Lists because an element can be made up of multiple solids. Each nested list represents one element solid.
    Does not work with in place elements.

    :param element: A revit element instance.
    :type element: Autodesk.Revit.DB.Element

    :return: A list of data geometry instances.
    :rtype: list of :class:`.DataGeometry`
    """

    all_element_points = []
    # get geometry from element
    opt = Options()
    fr1_geom = element.get_Geometry(opt)
    solids = []
    # check geometry for Solid elements
    # todo check for FamilyInstance geometry ( in place families!)
    for item in fr1_geom:
        if type(item) is Solid:
            solids.append(item)

    # process solids to points
    # in place families may have more then one solid
    for s in solids:
        points_per_solid = convert_solid_to_flattened_2d_points(s)
        if len(points_per_solid) > 0:
            for points_lists in points_per_solid:
                all_element_points.append(points_lists)
    return all_element_points


def get_solid_bounding_box(solid):
    """
    Returns a bounding box from a solid.

    :param solid: The solid to get the bounding box from.
    :type solid: Autodesk.Revit.DB.Solid

    :return: The bounding box of the solid.
    :rtype: Autodesk.Revit.DB.BoundingBoxXYZ
    """

    # get the solids bounding box
    solid_bounding_box = solid.GetBoundingBox()

    # transform the bounding box to the solids transform
    # which is different from the family instance transform!!
    solid_transform_min = solid_bounding_box.Transform.OfPoint(solid_bounding_box.Min)
    solid_transform_max = solid_bounding_box.Transform.OfPoint(solid_bounding_box.Max)

    # create a new bounding box from the transformed points
    solid_transform_bb = BoundingBoxXYZ()
    solid_transform_bb.Min = solid_transform_min
    solid_transform_bb.Max = solid_transform_max

    return solid_transform_bb


def get_bounding_box_from_family_geometry(geometry_element):
    """
    Returns a bounding box from the families solid elements geometry only.
    This is different from the family instance bounding box!

    :param geometry_element: The geometry element of a family instance.
    :type geometry_element: Autodesk.Revit.DB.GeometryElement

    :return: The bounding box of the family geometry.
    :rtype: Autodesk.Revit.DB.BoundingBoxXYZ
    """

    merged_result = None
    for geometry_obj in geometry_element:
        if geometry_obj is not None:
            # Instance geometry can also be a Solid
            if isinstance(geometry_obj, Solid):
                return get_solid_bounding_box(instance_geometry)
            # If not a solid, it is a list of geometry objects
            instance_geometry = geometry_obj.GetInstanceGeometry()
            if instance_geometry is not None:
                for element in instance_geometry:
                    # find solids
                    if type(element) is Solid:
                        # check if solid is valid
                        if element.Id == ElementId.InvalidElementId.IntegerValue:
                            continue
                        # get the solids bounding box
                        solid_transform_bb = get_solid_bounding_box(element)

                        # check if this is the first bounding box
                        if merged_result == None:
                            merged_result = solid_transform_bb
                            continue

                        # merge the bounding boxes
                        merged_result = merge_bounding_box_xyz(
                            merged_result, solid_transform_bb
                        )

    # return the merged bounding box
    return merged_result



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
This module contains the Revit view template data to 3D hash report functionality.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

from duHast.Utilities.Objects import result as res
from duHast.Revit.Views.Reporting.views_data_report import read_view_data_from_file
from duHast.Utilities.files_get import (
    get_file_name_without_ext,
)
from duHast.Revit.Views.Reporting.views_data_3d_hash_report_utilities import (
    _get_hash_headers,
    _get_hash_rows_categories,
    _get_hash_rows_filters,
    _get_hash_for_category_overrides,
    _get_hash_for_filter_overrides,
)
from duHast.Revit.Views.Reporting.view_reports_json_props import PROP_FILE_NAME

from duHast.Revit.Views.Reporting.Objects.json_conversion_storage import (
    JSONThreeDStorage,
)


def _load_json_data(files, progress_call_back=None):
    """
    Reads view template data from files into a dictionary.

    :param files: List of fully qualified file path to json files containing view template data.
    :type files: [str]
    :param progress_call_back: A call back function accepting as arguments the number of the current file processed and the number of overall files to be processed, defaults to None
    :type progress_call_back: func(counter, overall_counter), optional
    :return: A dictionary where key is the file name without extension and value is a list of view settings
    :rtype: {str: [:class:`.ViewGraphicsSettings`]}
    """

    json_data = {}
    counter = 0
    for file_path in files:
        file_name = get_file_name_without_ext(file_path=file_path)
        json_single_data = read_view_data_from_file(file_path=file_path)
        json_data[file_name] = json_single_data
        counter = counter + 1
        if progress_call_back is not None:
            progress_call_back(counter, len(files))
    return json_data


def _get_category_hash_table_data_by_file(view_settings, progress_call_back=None):
    """
    Returns a dictionary of partly populated JSONThreeDStorage objects.

    - column headers ( template names)
    - row headers ( category names)
    - hash table ( category overwrites )

    :param view_settings: A dictionary where key is the file name without extension and value is a list of view settings
    :type view_settings: {str: [:class:`.ViewGraphicsSettings`]}
    :param progress_call_back: A call back function accepting as arguments the number of the current file processed and the number of overall files to be processed, defaults to None
    :type progress_call_back: func(counter, overall_counter), optional
    :return: A dictionary where key is the file name without extension and value is an instance of a custom storage object
    :rtype: {str: [:class:`.JSONThreeDStorage`]}
    """

    dic_tables_by_file = {}
    counter = 0
    for key, vt_setting in view_settings.items():
        column_headers = _get_hash_headers(vt_setting)
        row_headers = _get_hash_rows_categories(vt_setting)
        row_headers_filters = _get_hash_rows_filters(vt_setting)
        hash_table_category_overrides = _get_hash_for_category_overrides(
            headers=column_headers, row_headers=row_headers, views_settings=vt_setting
        )
        hash_table_filter_overrides = _get_hash_for_filter_overrides(
            headers=column_headers,
            row_headers=row_headers_filters,
            views_settings=vt_setting,
        )
        storage = JSONThreeDStorage()
        storage.column_headers = column_headers
        storage.row_headers = row_headers
        storage.row_headers_filters = row_headers_filters
        storage.hash_table = hash_table_category_overrides
        storage.hash_table_filters = hash_table_filter_overrides
        dic_tables_by_file[key] = storage
        counter = counter + 1
        if progress_call_back is not None:
            progress_call_back(counter, len(view_settings))

    return dic_tables_by_file


def _map_hash_values_to_range(hash_data_by_file, progress_call_back=None):
    """
    Hash values returned from view templates have a really large range. This function maps them to their index in a sorted list.
    Note -1,0,1 are left unchanged.
    Hash tables for category and filter overrides will be updated with mapped values.

    :param hash_data_by_file: A dictionary where key is the file name without extension and value is an instance of a custom storage object
    :type hash_data_by_file: {str: [:class:`.JSONThreeDStorage`]}
    :param progress_call_back: A call back function accepting as arguments the number of the current file processed and the number of overall files to be processed, defaults to None
    :type progress_call_back: func(counter, overall_counter), optional
    :return: Passed in dictionary with updated hash tables.
    :rtype: {str: [:class:`.JSONThreeDStorage`]}
    """

    hash_mapper_categories = []
    hash_mapper_filters = []
    try:
        # build a unique list of all hash values
        for key, hash_data in hash_data_by_file.items():
            for entry in hash_data.hash_table:
                hash_mapper_categories = sorted(
                    list(set(hash_mapper_categories) | set(entry))
                )
            for entry in hash_data.hash_table_filters:
                hash_mapper_filters = sorted(
                    list(set(hash_mapper_filters) | set(entry))
                )

        # note 0, 1, -1 are special values and can not need to be changed!!
        # remove them from the sorted list and then re-insert them at the beginning
        if -1 in hash_mapper_categories:
            index = hash_mapper_categories.index(-1)
            hash_mapper_categories.pop(index)
        if 0 in hash_mapper_categories:
            index = hash_mapper_categories.index(0)
            hash_mapper_categories.pop(index)
        if 1 in hash_mapper_categories:
            index = hash_mapper_categories.index(1)
            hash_mapper_categories.pop(index)

        hash_mapper_categories.insert(0, -1)
        hash_mapper_categories.insert(0, 0)
        hash_mapper_categories.insert(0, 1)

        if -1 in hash_mapper_filters:
            index = hash_mapper_filters.index(-1)
            hash_mapper_filters.pop(index)
        if 0 in hash_mapper_filters:
            index = hash_mapper_filters.index(0)
            hash_mapper_filters.pop(index)
        if 1 in hash_mapper_filters:
            index = hash_mapper_filters.index(1)
            hash_mapper_filters.pop(index)

        hash_mapper_filters.insert(0, -1)
        hash_mapper_filters.insert(0, 0)
        hash_mapper_filters.insert(0, 1)

        # loop over current hash values and replace with index of value in mapper list
        # preserve -1, 0, 1 values

        call_back_progress_counter = 0
        for key, hash_data in hash_data_by_file.items():
            # update categories hash table
            mapped_hash_table_categories = []
            try:
                for row in hash_data.hash_table:
                    new_row = []
                    for entry in row:
                        if entry > 1 or entry < -1:
                            mapped_index = hash_mapper_categories.index(entry)
                            new_row.append(mapped_index)
                        else:
                            new_row.append(entry)
                    mapped_hash_table_categories.append(new_row)
                # overwrite existing hash table
                hash_data.hash_table = mapped_hash_table_categories
            except Exception as e:
                raise ValueError ("Failed to match category hash: {}".format(e))
            # update filters hash table
            mapped_hash_table_filters = []
            try:
                for row in hash_data.hash_table_filters:
                    new_row = []
                    for entry in row:
                        if entry > 1 or entry < -1:
                            mapped_index = hash_mapper_filters.index(entry)
                            new_row.append(mapped_index)
                        else:
                            new_row.append(entry)
                    mapped_hash_table_filters.append(new_row)
                # overwrite existing hash table
                hash_data.hash_table_filters = mapped_hash_table_filters
            except Exception as e:
                raise ValueError ("Failed to match filter hash: {}".format(e))
            # update call back
            call_back_progress_counter = call_back_progress_counter + 1
            if progress_call_back is not None:
                progress_call_back(call_back_progress_counter, len(hash_data_by_file))
    except Exception as e:
        raise ValueError("Failed to map values to range: {}".format(e))
    return hash_data_by_file


def _merge_column_headers(hash_data_by_file):
    """
    Builds a unique list of column headers ( view template names )
    This list is the same for filter and category overrides.

    :param hash_data_by_file: A dictionary where key is the file name without extension and value is an instance of a custom storage object
    :type hash_data_by_file: {str: [:class:`.JSONThreeDStorage`]}
    :return: Passed in dictionary with updated merged column list.
    :rtype: {str: [:class:`.JSONThreeDStorage`]}
    """

    try:
        data = []
        # build list with unique entries
        for key, vt_setting in hash_data_by_file.items():
            data = sorted(list(set(data) | set(vt_setting.column_headers)))

        # this list is common for all files...update them
        for key, vt_setting in hash_data_by_file.items():
            vt_setting.merged_column_headers = data

        return hash_data_by_file

    except Exception as e:
        raise ValueError(
            "Failed to merge column headers (view template names): {}".format(e)
        )


def _merge_row_headers(hash_data_by_file):
    """
    Builds a unique list of row headers for  object categories names and filter names.

    :param hash_data_by_file: A dictionary where key is the file name without extension and value is an instance of a custom storage object
    :type hash_data_by_file: {str: [:class:`.JSONThreeDStorage`]}
    :return: Passed in dictionary with updated merged row list.
    :rtype: {str: [:class:`.JSONThreeDStorage`]}
    """

    try:
        data_category_names = []
        data_filter_names = []
        # build list with unique entries for category names
        for key, vt_setting in hash_data_by_file.items():
            data_category_names = sorted(
                list(set(data_category_names) | set(vt_setting.row_headers))
            )

        # build list with unique entries for filter names
        for key, vt_setting in hash_data_by_file.items():
            data_filter_names = sorted(
                list(set(data_filter_names) | set(vt_setting.row_headers_filters))
            )

        # these lists are common for all files...update them
        for key, vt_setting in hash_data_by_file.items():
            vt_setting.merged_row_headers = data_category_names
            vt_setting.merged_row_headers_filters = data_filter_names

        return hash_data_by_file

    except Exception as e:
        raise ValueError(
            "Failed to merge row headers (object category names): {}".format(e)
        )


# ---------------------------- padded default array -------------------------


def _get_padded_default_array(
    merged_row_headers,
    merged_column_headers,
):
    """
    Builds an 2d array where column number equals the number of unique view templates (merged column header)
    and row number equals number of unique object category names. All values are set to -1.

    :param merged_row_headers: A list containing row headers
    :type merged_row_headers: [str]
    :param merged_column_headers: A list containing column headers
    :type merged_column_headers: [str]
    :return: A 2D array where all values are -1.
    :rtype: [[],[],...]
    """

    # Create a new padded 2D array
    padded_array = [
        [-1 for entry in merged_column_headers] for entry in merged_row_headers
    ]
    return padded_array


def _assign_padded_default_array(hash_data_by_file, progress_call_back=None):
    """
    Assigns a padded 2D array to storage class property

    :param hash_data_by_file: A dictionary where key is the file name without extension and value is an instance of a custom storage object
    :type hash_data_by_file: {str: [:class:`.JSONThreeDStorage`]}
    :param progress_call_back: A call back function accepting as arguments the number of the current file processed and the number of overall files to be processed, defaults to None
    :type progress_call_back: func(counter, overall_counter), optional
    :return:
        Result class instance.
        - .status True if padded array was successfully assigned. Otherwise False.
        - .message will contain array size.
    :rtype: :class:`.Result`
    """

    result = res.Result()
    try:
        call_back_progress_counter = 0
        # build default hash tables where all values are -1 for categories and filters
        for key, hash_by_file in hash_data_by_file.items():
            padded_array_categories = _get_padded_default_array(
                merged_column_headers=hash_by_file.merged_column_headers,
                merged_row_headers=hash_by_file.merged_row_headers,
            )
            padded_array_filters = _get_padded_default_array(
                merged_column_headers=hash_by_file.merged_column_headers,
                merged_row_headers=hash_by_file.merged_row_headers_filters,
            )
            hash_data_by_file[key].padded_default_hash_table = padded_array_categories
            hash_data_by_file[key].padded_default_hash_table_filters = padded_array_filters
            result.append_message(
                "{}: Created padded categories default value array of size {} by {}".format(
                    key, len(padded_array_categories), len(padded_array_categories[0])
                )
            )
            result.append_message(
                "{}: Created padded filters default value array of size {} by {}".format(
                    key, len(padded_array_filters), len(padded_array_filters[0])
                )
            )
            call_back_progress_counter = call_back_progress_counter + 1
            if progress_call_back is not None:
                progress_call_back(call_back_progress_counter, len(hash_data_by_file))
        result.result.append(hash_data_by_file)
    except Exception as e:
        result.update_sep(
            False, "Failed to assigned padded default array with: {}".format(e)
        )
    return result


# ---------------------------- row indices ---------------------------------


def _assign_row_indices_pointer(hash_data_by_file, progress_call_back=None):
    """
    Creates row and index pointers per file which map the file specific rows and column to the overall padded array rows and columns.

    :param hash_data_by_file: A dictionary where key is the file name without extension and value is an instance of a custom storage object
    :type hash_data_by_file: {str: [:class:`.JSONThreeDStorage`]}
    :param progress_call_back: A call back function accepting as arguments the number of the current file processed and the number of overall files to be processed, defaults to None
    :type progress_call_back: func(counter, overall_counter), optional
    :return:
        Result class instance.
        - .status True if row and column index pointers where successfully assigned. Otherwise False.
        - .message will contain array size.
    :rtype: :class:`.Result`
    """

    result = res.Result()
    try:
        call_back_progress_counter = 0
        # build row and column indices list for mapping of value hash table entries to default hash table
        for key, hash_by_file in hash_data_by_file.items():
            # Find the indices for row and column headers in the merged headers
            row_indices_categories_all = [
                hash_by_file.merged_row_headers.index(row)
                for row in hash_by_file.row_headers
            ]
            row_indices_filters_all = [
                hash_by_file.merged_row_headers_filters.index(row)
                for row in hash_by_file.row_headers_filters
            ]
            # column header indices are the same for filters and categories
            column_indices_all = [
                hash_by_file.merged_column_headers.index(col)
                for col in hash_by_file.column_headers
            ]

            hash_data_by_file[key].row_indices = row_indices_categories_all
            hash_data_by_file[key].row_indices_filters = row_indices_filters_all
            hash_data_by_file[key].column_indices = column_indices_all
            result.append_message(
                "{}: Created categories row: {} , filters row {} and column: {} indices mapper.".format(
                    key,
                    len(row_indices_categories_all),
                    len(row_indices_filters_all),
                    len(column_indices_all),
                )
            )
            call_back_progress_counter = call_back_progress_counter + 1
            if progress_call_back is not None:
                progress_call_back(call_back_progress_counter, len(hash_data_by_file))
        result.result.append(hash_data_by_file)
    except Exception as e:
        result.update_sep(
            False, "Failed to assigned row and column index pointers with: {}".format(e)
        )
    return result


def _update_default_array_values(row_indices, col_indices, default_array, value_array):
    """
    Updates the default padded arrays with hash values from file specific array using row and column index pointers.

    :param row_indices: Row index mapper from file specific array to default array.
    :type row_indices: [int]
    :param col_indices: _Column index mapper from file specific array to default array.
    :type col_indices: [int]
    :param default_array: The default 2D array
    :type default_array: [[int],[int],]
    :param value_array: The file specific 2D array
    :type value_array: [[int],[int],]
    :return: Default array updated with file specific array values.
    :rtype: [[int],[int],]
    """

    # Fill in the values from array_model_a
    for i, row_index in enumerate(row_indices):
        for j, col_index in enumerate(col_indices):
            default_array[row_index][col_index] = value_array[i][j]
    return default_array


def _assign_default_array_values(hash_data_by_file, progress_call_back=None):
    """
    Assigns file specific array values to default arrays for categories and filters by file.

    :param hash_data_by_file: A dictionary where key is the file name without extension and value is an instance of a custom storage object
    :type hash_data_by_file: {str: [:class:`.JSONThreeDStorage`]}
    :param progress_call_back: A call back function accepting as arguments the number of the current file processed and the number of overall files to be processed, defaults to None
    :type progress_call_back: func(counter, overall_counter), optional
    :return:
        Result class instance.
        - .status True if default arrays where successfully updated with values from file specific arrays. Otherwise False.
        - .message will contain array size.
    :rtype: :class:`.Result`
    """

    result = res.Result()
    try:
        call_back_progress_counter = 0
        # update the default hash table for each file with values from the value hash table from the same file
        for key, hash_by_file in hash_data_by_file.items():
            updated_array_categories = _update_default_array_values(
                row_indices=hash_by_file.row_indices,
                col_indices=hash_by_file.column_indices,
                default_array=hash_by_file.padded_default_hash_table,
                value_array=hash_by_file.hash_table,
            )

            updated_array_filters = _update_default_array_values(
                row_indices=hash_by_file.row_indices_filters,
                col_indices=hash_by_file.column_indices,
                default_array=hash_by_file.padded_default_hash_table_filters,
                value_array=hash_by_file.hash_table_filters,
            )

            hash_by_file.padded_value_hash_table = updated_array_categories
            hash_by_file.padded_value_hash_table_filters = updated_array_filters

            result.append_message(
                "{}: Updated padded category default value array with values of size {} by {}".format(
                    key,
                    len(updated_array_categories),
                    len(updated_array_categories[0]),
                )
            )

            result.append_message(
                "{}: Updated padded filter default value array with values of size {} by {}".format(
                    key,
                    len(updated_array_filters),
                    len(updated_array_filters[0]),
                )
            )

            call_back_progress_counter = call_back_progress_counter + 1
            if progress_call_back is not None:
                progress_call_back(call_back_progress_counter, len(hash_data_by_file))
        result.result.append(hash_data_by_file)
    except Exception as e:
        result.update_sep(
            False,
            "Failed to assigned file specific array values to default array: {}".format(
                e
            ),
        )
    return result


def _built_threeD_array(hash_data_by_file):
    """
    Combine 2D default hash arrays from all files into single 3D hash array.

    :param hash_data_by_file: A dictionary where key is the file name without extension and value is an instance of a custom storage object
    :type hash_data_by_file: {str: [:class:`.JSONThreeDStorage`]}
    :return:
        Result class instance.
        - .status True if 3D array was successfully created. Otherwise False.
        - .message will contain array size.
        -. result will contain the 3D array for categories as first value in list and the 3D array for filters as the second value in list
    :rtype: :class:`.Result`
    """

    result = res.Result()
    try:
        array_3d_categories = []

        z_value_categories = 0
        x_value_categories = 0
        y_value_categories = 0

        array_3d_filters = []
        z_value_filters = 0
        x_value_filters = 0
        y_value_filters = 0

        for key, hash_by_file in hash_data_by_file.items():
            # categories
            array_3d_categories.append(hash_by_file.padded_value_hash_table)
            x_value_categories = len(hash_by_file.padded_value_hash_table)
            y_value_categories = len(hash_by_file.padded_value_hash_table[0])
            z_value_categories = z_value_categories + 1
            # filters
            array_3d_filters.append(hash_by_file.padded_value_hash_table_filters)
            x_value_filters = len(hash_by_file.padded_value_hash_table_filters)
            y_value_filters = len(hash_by_file.padded_value_hash_table_filters[0])
            z_value_filters = z_value_filters + 1

        result.append_message(
            "Built 3D category array of size: number of files: {} , by number of categories: {} , by number of view templates: {}".format(
                z_value_categories, x_value_categories, y_value_categories
            )
        )
        result.append_message(
            "Built 3D filter array of size: number of files: {} , by number of filters: {} , by number of view templates: {}".format(
                z_value_filters, x_value_filters, y_value_filters
            )
        )

        result.result.append(array_3d_categories)
        result.result.append(array_3d_filters)
    except Exception as e:
        result.update_sep(False, "Failed to build 3D array: {}".format(e))
    return result


def _flatten_category_threeD_array(
    array_3d, sample_storage, model_names, hash_data_by_file, progress_call_back=None
):
    """
    Flattens the 3D array build from hash tables from each file into a json structure easily read by power bi.

    :param array_3d: A 3D array.
    :type array_3d: [[[]]]
    :param sample_storage: Storage instance used to map view template names and category names using merged lists.
    :type sample_storage: :class:`.JSONThreeDStorage`
    :param model_names: List of all model names of which view template data is included in 3D hash array.
    :type model_names: [str]
    :param hash_data_by_file: A dictionary where key is the file name without extension and value is an instance of a custom storage object
    :type hash_data_by_file: {str: [:class:`.JSONThreeDStorage`]}
    :param progress_call_back: A call back function accepting as arguments the number of the current file processed and the number of overall files to be processed, defaults to None
    :type progress_call_back: func(counter, overall_counter), optional
    :return:
        Result class instance.
        - .status True if 3D array was successfully created. Otherwise False.
        - .message will contain array size.
        -. result will contain the flatten array as first value in list
    :rtype: :class:`.Result`
    """

    result = res.Result()
    try:
        call_back_progress_counter = 0
        # flatten 3D hash data for power bi
        flattened_data = []
        for model_name, layer in enumerate(array_3d):
            for category, row in enumerate(layer):
                for view_template, hash_value in enumerate(row):
                    flattened_data.append(
                        {
                            "view_template": sample_storage.merged_column_headers[
                                view_template
                            ],
                            "category": sample_storage.merged_row_headers[category],
                            "model_name": model_names[model_name],
                            "hash_value": hash_value,
                        }
                    )
            call_back_progress_counter = call_back_progress_counter + 1
            if progress_call_back is not None:
                progress_call_back(call_back_progress_counter, len(hash_data_by_file))
        result.result.append(flattened_data)
    except Exception as e:
        result.update_sep(False, "Failed to flatten 3D array: {}".format(e))
    return result


def _flatten_filter_threeD_array(
    array_3d, sample_storage, model_names, hash_data_by_file, progress_call_back=None
):
    """
    Flattens the 3D array build from hash tables from each file into a json structure easily read by power bi.

    :param array_3d: A 3D array.
    :type array_3d: [[[]]]
    :param sample_storage: Storage instance used to map view template names and category names using merged lists.
    :type sample_storage: :class:`.JSONThreeDStorage`
    :param model_names: List of all model names of which view template data is included in 3D hash array.
    :type model_names: [str]
    :param hash_data_by_file: A dictionary where key is the file name without extension and value is an instance of a custom storage object
    :type hash_data_by_file: {str: [:class:`.JSONThreeDStorage`]}
    :param progress_call_back: A call back function accepting as arguments the number of the current file processed and the number of overall files to be processed, defaults to None
    :type progress_call_back: func(counter, overall_counter), optional
    :return:
        Result class instance.
        - .status True if 3D array was successfully created. Otherwise False.
        - .message will contain array size.
        -. result will contain the flatten array as first value in list
    :rtype: :class:`.Result`
    """

    result = res.Result()
    try:
        call_back_progress_counter = 0
        # flatten 3D hash data for power bi
        flattened_data = []
        for model_name, layer in enumerate(array_3d):
            for filter, row in enumerate(layer):
                for view_template, hash_value in enumerate(row):
                    flattened_data.append(
                        {
                            "view_template": sample_storage.merged_column_headers[
                                view_template
                            ],
                            "filter": sample_storage.merged_row_headers_filters[filter],
                            "model_name": model_names[model_name],
                            "hash_value": hash_value,
                        }
                    )
            call_back_progress_counter = call_back_progress_counter + 1
            if progress_call_back is not None:
                progress_call_back(call_back_progress_counter, len(hash_data_by_file))
        result.result.append(flattened_data)
    except Exception as e:
        result.update_sep(False, "Failed to flatten 3D array: {}".format(e))
    return result


def convert_vt_data_to_3d_flattened(json_files, progress_call_back=None):
    """
    Converts view template graphic overrides data stored in files into flattened json formatted hash table array for import to power bi.

    :param json_files: List of files containing view template data of Revit project files. ( One json file per Revit project file)
    :type json_files: [str]
    :param progress_call_back: A call back function accepting as arguments the number of the current file processed and the number of overall files to be processed, defaults to None
    :type progress_call_back: func(counter, overall_counter), optional

    :return:
        Result class instance.
        - .status True if flattened array was successfully created. Otherwise False.
        - .message will contain array size.
        -. result will contain the flatten array as first value in list
    :rtype: :class:`.Result`
    """

    result = res.Result()
    try:
        # load json data from all files
        json_data_loaded = _load_json_data(
            files=json_files, progress_call_back=progress_call_back
        )
        result.append_message(
            "Loaded json data from {} files.".format(len(json_data_loaded))
        )

        # get hash tables, row and column data, key is the file name
        hash_data_by_file = _get_category_hash_table_data_by_file(
            json_data_loaded, progress_call_back=progress_call_back
        )

        # map hash values to a range
        hash_data_by_file = _map_hash_values_to_range(hash_data_by_file)

        # merge row and column headers into unique value lists and store them in each storage instance
        hash_data_by_file = _merge_column_headers(hash_data_by_file=hash_data_by_file)
        hash_data_by_file = _merge_row_headers(hash_data_by_file=hash_data_by_file)

        for key, entry in hash_data_by_file.items():
            result.append_message(
                "{}: number of merged column headers: {} , merged category row headers: {} and number of merged filter row headers: {}".format(
                    key, len(entry.merged_column_headers), len(entry.merged_row_headers), len(entry.merged_row_headers_filters)
                )
            )

        # assign padded default hash table
        assign_padded_default_hash_table_status = _assign_padded_default_array(
            hash_data_by_file=hash_data_by_file, progress_call_back=progress_call_back
        )
        result.update(assign_padded_default_hash_table_status)
        # check if all is ok
        if assign_padded_default_hash_table_status.status:
            # get the updated storage items
            hash_data_by_file = assign_padded_default_hash_table_status.result[0]
            # get row and column indices mapping to padded hash table from value table
            row_indices_status = _assign_row_indices_pointer(
                hash_data_by_file, progress_call_back=progress_call_back
            )
            result.update(row_indices_status)
            # check if all is ok
            if row_indices_status.status:
                # get the updated storage items
                hash_data_by_file = row_indices_status.result[0]
                # update the padded hash table of default values with values from the actual hash table mapped to columns and rows
                assign_default_array_values_status = _assign_default_array_values(
                    hash_data_by_file, progress_call_back=progress_call_back
                )
                result.update(assign_default_array_values_status)
                # check if all is ok
                if assign_default_array_values_status.status:
                    # get the updated storage items
                    hash_data_by_file = assign_default_array_values_status.result[0]

                    # build an 3D array from padded hash value tables
                    array_3d_status = _built_threeD_array(hash_data_by_file)
                    # check if all is ok
                    if array_3d_status.status:
                        # get the 3D array
                        array_3d_categories = array_3d_status.result[0]
                        array_3d_filters = array_3d_status.result[1]

                        # Get an arbitrary key-value pair using the dictionary's iterator
                        # since all storage instance store the same merged rows and column headers lists
                        first_key, first_value = next(iter(hash_data_by_file.items()))

                        # built a list of model names from the keys of the hash data dictionary
                        model_names = list(hash_data_by_file.keys())

                        # flatten the category 3D array for power bi
                        flatten_category_array_status = _flatten_category_threeD_array(
                            array_3d=array_3d_categories,
                            sample_storage=first_value,
                            model_names=model_names,
                            hash_data_by_file=hash_data_by_file,
                            progress_call_back=progress_call_back,
                        )
                        result.update(flatten_category_array_status)

                        # flatten the filter 3D array for power bi
                        flatten_filter_array_status = _flatten_filter_threeD_array(
                            array_3d=array_3d_filters,
                            sample_storage=first_value,
                            model_names=model_names,
                            hash_data_by_file=hash_data_by_file,
                            progress_call_back=progress_call_back,
                        )
                        result.update(flatten_filter_array_status)

                        # check if all is ok
                        if (
                            flatten_category_array_status.status
                            and flatten_filter_array_status.status
                        ):
                            # get the flattened 3D arrays
                            result.result = [
                                flatten_category_array_status.result[0],
                                flatten_filter_array_status.result[0],
                            ]
                        else:
                            raise ValueError(
                                "Failed to flatten 3D arrays: {} and {}".format(
                                    flatten_category_array_status.message,
                                    flatten_filter_array_status.message,
                                )
                            )
                    else:
                        raise ValueError(
                            "Failed to build 3D arrays: {}".format(
                                array_3d_status.message
                            )
                        )
                else:
                    raise ValueError(
                        "Failed to update values in default hash table with values from source table: {}".format(
                            assign_default_array_values_status.message
                        )
                    )
            else:
                raise ValueError(
                    "Failed to assign row and column indices: {}".format(
                        row_indices_status.message
                    )
                )
        else:
            raise ValueError(
                "Failed to assign padded default hash table: {}".format(
                    assign_padded_default_hash_table_status.message
                )
            )
    except Exception as e:
        result.update_sep(
            status=False,
            message="Failed to build flattened 3D array with: {}".format(e),
        )
    return result



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
A parameter directive base class
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Parameter directives are used to modify parameters of elements in a Revit model. This class is the base class for all parameter directives.

"""

# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed. 
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits; 
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

import json
from duHast.Utilities.Objects import base


class ParameterDirectiveBase(base.Base):
    """Base class for all parameter directives."""

    data_type = "parameter_directive_base"

    def __init__(
        self,
        target_parameter_name="",
        parameter_modifier=None,
        parameter_value=None,
        j=None,
        **kwargs
    ):
        """
        Constructor for the ParameterDirective class.

        :param target_parameter_name: The name of the parameter to be modified.
        :type target_parameter_name: str
        :param parameter_modifier: The modifier to be applied to the parameter value.
        :type parameter_modifier: A function accepting the value as an argument and returning the modified value.
        :param parameter_value: The value to be applied to the parameter.
        :type parameter_value: any
        :param j: A json object containing the properties of the class.
        :type j: json
        :param kwargs: Additional keyword arguments.
        :type kwargs: dict

        :raises ValueError: If the target_parameter_name or parameter_value is not supplied in j.

        """
        super(ParameterDirectiveBase, self).__init__(**kwargs)

        self.target_parameter_name = target_parameter_name
        self.parameter_modifier = parameter_modifier
        self.parameter_value = parameter_value

        # check if any data was past in with constructor!
        if j != None and len(j) > 0:
            # check type of data that came in:
            if type(j) == str:
                # a string
                j = json.loads(j)
            elif type(j) == dict:
                # no action required
                pass
            else:
                raise ValueError(
                    "Argument supplied must be of type string or type dictionary"
                )

            # load values and throw exception if something is missing!
            # do not load parameter modifier and getter functions...
            try:
                self.target_parameter_name = j["target_parameter_name"]
                self.parameter_value = j["parameter_value"]
            except Exception as e:
                raise ValueError(
                    "Node {} failed to initialise with: {}".format(
                        ParameterDirectiveBase.data_type, e
                    )
                )

    def __eq__(self, other):
        """
        Custom compare is equal override. Ignores properties:

        - parameter_modifier

        :param other: Another instance of  ParameterDirectiveBase  class
        :type other: :class:`. ParameterDirectiveBase`
        :return: True if all properties of compared class instances are equal, otherwise False.
        :rtype: Bool
        """

        return isinstance(other, ParameterDirectiveBase) and (
            self.target_parameter_name,
            self.parameter_value,
        ) == (
            other.target_parameter_name,
            other.parameter_value,
        )

    # python 2.7 needs custom implementation of not equal
    def __ne__(self, other):
        return not self.__eq__(other=other)

    # overwrite the base class to_json function
    def to_json(self):
        """
        Convert the instance of this class to json.
        Ignores parameter modifier property

        :return: A Json object.
        :rtype: json
        """

        # properties excluded from json dump
        excluded_properties = [
            "parameter_modifier",
        ]

        return json.dumps(
            self,
            indent=None,
            default=lambda o: {
                key: value
                for key, value in o.__dict__.items()
                if key not in excluded_properties
            },
        )



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Family data collector class.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

- Collects data from current family document and then recursively from any nested family.

"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

from duHast.Utilities.Objects import result as res

# import Autodesk
from Autodesk.Revit.DB import Element, FamilySymbol, FilteredElementCollector
from duHast.Utilities.Objects import base


class RevitFamilyDataCollector(base.Base):
    def __init__(self, data_processors):
        """
        Class constructor taking a list of processor instances as argument.

        :param data_processors: List of processor instances
        :type data_processors: [IFamilyProcessor]
        """

        # forwards all unused arguments
        # ini super class to allow multi inheritance in children!
        super(RevitFamilyDataCollector, self).__init__()

        self.dataProcessors = data_processors

    def _get_family_ids(self, doc):
        """
        Get all loadable family ids in file.

        :param doc: Current family document.
        :type doc: Autodesk.Revit.DB.Document
        :return: list of family ids
        :rtype: [Autodesk.Revit.DB.ElementId]
        """

        family_ids = []
        col = FilteredElementCollector(doc).OfClass(FamilySymbol)
        # get families from symbols and filter out in place families
        for fam_symbol in col:
            if (
                fam_symbol.Family.Id not in family_ids
                and fam_symbol.Family.IsInPlace == False
            ):
                family_ids.append(fam_symbol.Family.Id)
        return family_ids

    def _dive(self, doc, root_name, root_category, is_root=False):
        """
        Loops recursively over each family nested into root family document and and calls processor instance\
             with the current family document.

        :param doc: The family document. 
        :type doc: Autodesk.Revit.DB.Document
        :param root_name: The path of the nested family in a tree: rootFamilyName::nestedFamilyNameOne::nestedFamilyTwo\
            This includes the actual family name as the last node.
        :type root_name: str
        :param is_root: Indicates whether document is that of the root family, defaults to False
        :type is_root: bool, optional
        """

        return_value = res.Result()
        # only process current doc if not the root family
        # that family is processed already
        if is_root == False:
            for pro in self.dataProcessors:
                try:
                    pro.process(doc, root_name, root_category)
                    return_value.append_message(
                        "Processor [{}] of family: {}  [OK]".format(
                            pro.data_type, doc.Title
                        )
                    )
                except Exception as e:
                    return_value.update_sep(
                        False,
                        "Processor [{}] of family: {} [EXCEPTION] {}".format(
                            pro.data_type, doc.Title, e
                        ),
                    )

        # check if family doc
        if doc.IsFamilyDocument:
            # get any nested families
            family_ids = self._get_family_ids(doc)
            # if there are any nested families open those for processing
            if len(family_ids) > 0:
                for family_id in family_ids:
                    family = doc.GetElement(family_id)
                    try:
                        if family.IsEditable and family.IsValidObject:
                            family_doc = doc.EditFamily(family)
                            fam_name = family.Name
                            # strip .rfa of name
                            if fam_name.lower().endswith(".rfa"):
                                fam_name = fam_name[:-4]
                            # get category
                            fam_category_name = family.FamilyCategory.Name
                            # go recursive
                            dive_result = self._dive(
                                family_doc,
                                root_name + " :: " + fam_name,
                                root_category + " :: " + fam_category_name,
                            )
                            return_value.update(dive_result)
                    except Exception as e:
                        message = ""
                        if family != None:
                            message = "An exception occurred when opening family {} . Exception: {}".format(
                                Element.Name.GetValue(family), e
                            )
                        else:
                            message = "An exception occurred when attempting the get family element by id: {}. Exception: {}".format(
                                family_id, e
                            )
                        return_value.update_sep(False, message)
            else:
                # only close any nested family document...not the root one
                # since that is closed by batch processor!
                if is_root == False:
                    try:
                        # no more families found. Close the active doc
                        doc.Close(False)
                    except Exception as e:
                        message = ""
                        if doc != None:
                            message = "An exception occurred when closing document: {} . Exception: {}".format(
                                doc.Title, e
                            )
                        else:
                            message = "An exception occurred when closing document: {}".format(
                                e
                            )
                        return_value.update_sep(False, message)
        return return_value

    def process_family(self, doc, root_name, root_category):
        """
        Entry point for recursive family looper.

        Processes root family as well as actions any post processing.

        :param doc: The family document. 
        :type doc: Autodesk.Revit.DB.Document
        :param root_name: The path of the nested family in a tree: rootFamilyName::nestedFamilyNameOne::nestedFamilyTwo\
            This includes the actual family name as the last node.
        :type root_name: str
        :param root_category: The path of the nested family category in a tree: rootFamilyCategory::nestedFamilyOneCategory::nestedFamilyTwoCategory\
            This includes the actual family category as the last node.
        :type root_category: str

        :return: 
            Result class instance.
            
            - .status True if all data processor instances ran without an exception. Otherwise False.
            - .message will contain each processor type and its processing status'
        
        :rtype: :class:`.Result`
        """

        return_value = res.Result()

        # TODO:
        # action any pre processing actions
        # loop over fam processor instances and process family with each of them
        for pro in self.dataProcessors:
            try:
                pre_action_result = pro.pre_process_actions(doc)
                return_value.update(pre_action_result)
            except Exception as e:
                return_value.update_sep(
                    False,
                    "PreProcessor [{}] of family: {} [EXCEPTION] {}".format(
                    pro.data_type,doc.Title,e)
                )

        # loop over fam processor instances and process family with each of them
        for pro in self.dataProcessors:
            try:
                pro.process(doc, root_name, root_category)
                return_value.append_message(
                    "Processor [{}] of family: {} [OK]".format(pro.data_type, doc.Title)
                )
            except Exception as e:
                return_value.update_sep(
                    False,
                    "Processor [{}] of family: {} [EXCEPTION] {}".format(
                        pro.data_type, doc.Title, e
                    ),
                )

        # check out any nested families
        dive_result = self._dive(doc, root_name, root_category, True)
        return_value.update(dive_result)

        # TODO:
        # action any post processing actions
        # loop over fam processor instances and process family with each of them
        for pro in self.dataProcessors:
            try:
                pro_action_result = pro.post_process_actions(doc)
                return_value.update(pro_action_result)
            except Exception as e:
                return_value.update_sep(
                    False,
                    "PostProcessor [{}] of family: {} [EXCEPTION] {}".format(
                        pro.data_type, doc.Title, e
                    ),
                )

        return return_value



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Helper functions to read rename directives file(s) and return them as a tuple to the caller
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

These helper function expect a text file in csv format with 4 columns:

- Current family name: with out the file extension
- File path	: fully qualified file path to the family file.
- Family category: the Revit category of the family.
- New family name: the new family name without the file extension.

Note:

- First row is treated as a header row and its content is ignored.

"""


#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed. 
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits; 
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

import clr
import System
from collections import namedtuple
from duHast.Utilities.Objects.timer import Timer

from duHast.Utilities import files_csv as fileCSV, files_get as fileGet
from duHast.Utilities.Objects import result as res
from duHast.Revit.Family.Data import family_base_data_utils as rFamBaseDUtils


# tuples containing rename directive read from file
rename_directive = namedtuple("rename_directive", "name filePath category newName")

# row structure of rename directive file

RENAME_DIRECTIVE_LIST_INDEX_CURRENT_FAMILY_NAME = 0
RENAME_DIRECTIVE_INDEX_FAMILY_FILE_PATH = 1
RENAME_DIRECTIVE_INDEX_CATEGORY = 2
RENAME_DIRECTIVE_LIST_INDEX_NEW_FAMILY_NAME = 3

# file name identifiers for rename directives
RENAME_DIRECTIVE_FILE_NAME_PREFIX = "RenameDirective"
RENAME_DIRECTIVE_FILE_EXTENSION = ".csv"

# exceptions
EXCEPTION_NO_RENAME_DIRECTIVE_FILES = "Rename directive file does not exist."
EXCEPTION_EMPTY_RENAME_DIRECTIVE_FILES = "Empty rename directive file!"


def _read_rename_directives(files):
    """
    Reads list of rename directives from file into named tuples.

    :param filePath: Fully qualified file path to rename directives file.
    :type filePath: str
    :return: List of named tuples containing rename directives.
    :rtype: [rename_directive]
    """

    rename_directives = []
    for file in files:
        rows = fileCSV.read_csv_file(file)
        # read rows in tuples ignoring the header row
        for i in range(1, len(rows)):
            if len(rows[i]) >= 4:
                data = rename_directive(
                    rows[i][RENAME_DIRECTIVE_LIST_INDEX_CURRENT_FAMILY_NAME],
                    rows[i][RENAME_DIRECTIVE_INDEX_FAMILY_FILE_PATH],
                    rows[i][RENAME_DIRECTIVE_INDEX_CATEGORY],
                    rows[i][RENAME_DIRECTIVE_LIST_INDEX_NEW_FAMILY_NAME],
                )
            rename_directives.append(data)
    return rename_directives


def get_rename_directives(directory_path):
    """
    Retrieves file rename  directives from a given folder location.

    :param directory_path: Fully qualified folder path to folder containing directives.
    :type directory_path: str

    :return:
        Result class instance.

        - result.status. True if rename directives where found and loaded successfully, otherwise False.
        - result.message will contain number of directives found in format:'Found rename directives: ' + number
        - result.result list of directives

        On exception:

        - result.status (bool) will be False.
        - result.message will contain an exception message.
        - result.result will be empty

    :rtype: :class:`.Result`
    """

    return_value = res.Result()
    # check whether csv files matching file name filter exist in directory path
    rename_directive_files = fileGet.get_files_from_directory_walker_with_filters(
        directory_path,
        RENAME_DIRECTIVE_FILE_NAME_PREFIX,
        "",
        RENAME_DIRECTIVE_FILE_EXTENSION,
    )

    # check whether any files where found?
    if len(rename_directive_files) > 0:
        # attempt to re rename directives from files
        rename_directives = _read_rename_directives(rename_directive_files)
        # check whether any rename directives where found in files
        if len(rename_directives) > 0:
            return_value.update_sep(
                True, "Found rename directives: " + str(len(rename_directives))
            )
            # attempt to rename files
            return_value.result = rename_directives
        else:
            return_value.update_sep(False, EXCEPTION_EMPTY_RENAME_DIRECTIVE_FILES)
    else:
        return_value.update_sep(False, EXCEPTION_NO_RENAME_DIRECTIVE_FILES)

    return return_value



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
This module contains a number of helper functions relating to Revit category overrides in views. 
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

from duHast.Revit.Views.Objects.category_override_storage import RevitCategoryOverride
from duHast.Utilities.Objects import result as res
from duHast.Revit.Common.transaction import in_transaction
from Autodesk.Revit.DB import Transaction


def get_categories_and_subcategories_from_model(doc):
    """
    Returns all categories and sub categories in a model in a dictionary:

    Dictionary:

        - key is same format as list: main category name::sub category name
        - value is a named tuple category_storage instance

    :param doc: The current model document.
    :type doc: Autodesk.Revit.DB.Document

    :return: dictionary of categories
    :rtype: {str:category_storage}
    """

    categories_dic = {}
    document_settings = doc.Settings
    cats = document_settings.Categories
    for main_category in cats:
        # add the main category
        categories_dic[
            "{}::{}".format(main_category.Name, main_category.Name)
        ] = RevitCategoryOverride(
            main_category_name=main_category.Name,
            sub_category_name=main_category.Name,
            category=main_category,
            category_id=main_category.Id,
            category_override=None,
            is_category_hidden=False,
        )
        # loop over the sub categories
        for sub_cat in main_category.SubCategories:
            categories_dic[
                "{}::{}".format(main_category.Name, sub_cat.Name)
            ] = RevitCategoryOverride(
                main_category_name=main_category.Name,
                sub_category_name=sub_cat.Name,
                category=sub_cat,
                category_id=sub_cat.Id,
                category_override=None,
                is_category_hidden=False,
            )

    return categories_dic


def update_category_override_from_view(view, category_storage_instance):
    """
    Populates the category_override and is_category_hidden field of a 'category_storage' instance based on the view past in.

    :param view: The view from which the category override is to be used.
    :type view: Autodesk.Revit.DB.View
    :param category_storage_instance: An instances of revit category storage containing no view specific information.
    :type category_storage_instance: :class:`.RevitCategoryOverride`

    :return: An updated instances of revit category storage now containing the override, category visibility flags.
    :rtype: :class:`.RevitCategoryOverride`
    """

    category_storage_instance.category_override = view.GetCategoryOverrides(
        category_storage_instance.category_id
    )
    category_storage_instance.is_category_hidden = view.GetCategoryHidden(
        category_storage_instance.category_id
    )

    return category_storage_instance


def get_category_overrides_from_view(view, category_storage_instances):
    """
    Populates the category_override and is_category_hidden fields of a list of 'category_storage' instances based on the view past in.

    :param view: The view from which the category override is to be used.
    :type view: Autodesk.Revit.DB.View
    :param category_storage_instances: A list of instances of class 'RevitCategoryOverride'
    :type category_storage_instances: [:class:`.RevitCategoryOverride`]

    :return: A list of instances of revit category storage now containing the override, category visibility flags.
    :rtype: [:class:`.RevitCategoryOverride`]
    """

    updated_category_storage_instances = []
    for category_storage_instance in category_storage_instances:
        updated_category_instance = update_category_override_from_view(
            view, category_storage_instance
        )
        updated_category_storage_instances.append(updated_category_instance)
    return updated_category_storage_instances


def apply_graphic_override_to_view(doc, view, category_storage_instances):
    """
    Applies graphic override(s) to a single view

    :param doc: The current model document.
    :type doc: Autodesk.Revit.DB.Document
    :param view: The view on which the category override is to be used.
    :type view: Autodesk.Revit.DB.View
    :param category_storage_instances: A list of instances of class 'RevitCategoryOverride'
    :type category_storage_instances: [:class:`.RevitCategoryOverride`]

    return:
        Result class instance.

        - Apply override status returned in result.status. False if an exception occurred, otherwise True.
        - result.message will contain message 'Successfully set category override...' for each override applied.
        - result.result will be an empty list

        On exception:

        - result.status (bool) will be False.
        - result.message will contain the exception message.
        - result.result will be an empty list.

    :rtype: :class:`.Result`
    """

    return_value = res.Result()
    for category_storage_instance in category_storage_instances:

        def action():
            action_return_value = res.Result()
            try:
                view.SetCategoryOverrides(
                    category_storage_instance.category_id,
                    category_storage_instance.revit_override,
                )
                action_return_value.update_sep(
                    True,
                    "Successfully set category override {} :: {} in view {} ".format(
                        category_storage_instance.main_category_name,
                        category_storage_instance.sub_category_name,
                        view.Name,
                    ),
                )
                view.SetCategoryHidden(
                    category_storage_instance.category_id,
                    category_storage_instance.is_category_hidden,
                )
                action_return_value.update_sep(
                    True,
                    "Successfully set category hidden status {} :: {} in view {} to: {}".format(
                        category_storage_instance.main_category_name,
                        category_storage_instance.sub_category_name,
                        view.Name,
                        category_storage_instance.is_category_hidden,
                    ),
                )
            except Exception as e:
                action_return_value.update_sep(
                    False,
                    "Failed to apply override: {} :: {} to view: {} with exception: {}".format(
                        category_storage_instance.main_category_name,
                        category_storage_instance.sub_category_name,
                        view.Name,
                        e,
                    ),
                )
            return action_return_value

        transaction = Transaction(
            doc,
            "Updating category override {} :: {}".format(
                category_storage_instance.main_category_name,
                category_storage_instance.sub_category_name,
            ),
        )
        update_category_override = in_transaction(transaction, action)
        return_value.update(update_category_override)

    return return_value



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Utility functions writing / reading json objects to/ from file.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed. 
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits; 
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

from duHast.Utilities.Objects import result as res

import codecs
import json

def _custom_default(o):
    '''
    Encode string values to utf-8 for json formatted outputs

    :param o: The value to be encoded if of type string
    :type o: var
    :return: Encoded string or var
    :rtype: str, var
    '''

    if isinstance(o, str):
        # only encode if required
        if any(ord(char) > 127 for char in o):
            return o.encode('utf-8').decode('utf-8')  # Encoding and decoding to ensure the type is str
    return o.__dict__

def write_json_to_file(json_data, data_output_file_path, enforce_utf8 = True):
    """
    Writes collected data to a new json formatted file.

    :param json_data: A dictionary to be written to file.
    :type json_data: json object (dictionary)
    :param data_output_file_path: Fully qualified file path to json data file.
    :type data_output_file_path: str
    :param enforce_utf8: Will encode any string value as utf-8, Default is true (recommended!!).
    :type enforce_utf8: bool
    :return:
        Result class instance.
        - result.status. True if json data file was written successfully, otherwise False.
        - result.message will confirm path of json data file.
        - result.result empty list
        On exception:
        - result.status (bool) will be False.
        - result.message will contain exception message.
        - result.result will be empty
    :rtype: :class:`.Result`
    """

    result = res.Result()

    try:
        json_object = None
        # check if utf-8 is to be enforced
        if(enforce_utf8):
            json_object = json.dumps(json_data, indent=None, default=_custom_default, ensure_ascii=False)
        else:
            json_object = json.dumps(json_data, indent=None, default=lambda o: o.__dict__)
        with codecs.open(data_output_file_path, "w", encoding="utf-8") as f:
            f.write(json_object)
            f.close()

        result.update_sep(
            True, "Data written to file: {}".format(data_output_file_path)
        )
    except Exception as e:
        result.update_sep(
            False, "Failed to write data to file with exception: {}".format(e)
        )
    return result


def read_json_data_from_file(file_path):
    """
    Reads json from file

    :param revit_file_path: Fully qualified file path of report file.
    :type file_path: str
    :return: json object
    :rtype: {}
    """

    data = {}
    try:
        # Opening JSON file
        with open(file_path) as f:
            # returns JSON object as
            # a dictionary
            data = json.load(f)
            f.close()
    except Exception as e:
        pass
    return data



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
This module contains a number of helper functions relating to Revit view filters. 
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed. 
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits; 
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

from Autodesk.Revit.DB import (
    Element,
    FilteredElementCollector,
    ParameterFilterElement,
    Transaction,
    View,
    ViewType,
)

from duHast.Revit.Views.templates import get_template_ids_which_can_have_filters
from duHast.Revit.Common import common as com
from duHast.Utilities.Objects import result as res
from duHast.Revit.Common.transaction import in_transaction


VIEW_TYPE_WHICH_CAN_HAVE_FILTERS = [
    ViewType.FloorPlan,
    ViewType.CeilingPlan,
    ViewType.Elevation,
    ViewType.ThreeD,
    ViewType.EngineeringPlan,
    ViewType.AreaPlan,
    ViewType.Section,
    ViewType.Detail,
    ViewType.Walkthrough,
    ViewType.DraftingView,
    ViewType.Legend,
]


def get_all_filters(doc):
    """
    Gets all filters in document as a collector

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: A filtered Element collector containing Autodesk.Revit.DB.ParameterFilterElement
    :rtype: Autodesk.Revit.DB.FilteredElementCollector
    """
    collector = FilteredElementCollector(doc).OfClass(ParameterFilterElement)
    return collector


def get_all_filter_ids(doc):
    """
    Gets all view filter ids in document

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: All view filter Id's which are in the model.
    :rtype: list of Autodesk.Revit.DB.ElementId
    """

    ids = []
    col = get_all_filters(doc)
    ids = com.get_ids_from_element_collector(col)
    return ids


def get_filter_by_name(doc, filter_name):
    """
    Retrieves a filter by its name from the current Revit model document.

    Args:
        doc (Revit Document): The current Revit model document.
        filter_name (str): The name of the filter to be retrieved.

    Returns:
        Filter: The filter with the specified name, if found.
        None: If no filter with the specified name is found.
    """
    filters = get_all_filters(doc=doc)
    for filter in filters:
        if Element.Name.GetValue(filter) == filter_name:
            return filter
    return None


def get_filter_ids_from_view_by_filter(view, unique_list):
    """
    Returns past in list of filter id's plus new unique filter id's from view (if not already in list past in)

    :param view: The view of which to get the filters from.
    :type view: Autodesk.Revit.DB.View
    :param unique_list: List containing view filters
    :type unique_list: list of Autodesk.Revit.DB.ElementId
    :return: List containing past in view filters and new view filters.
    :rtype: list of Autodesk.Revit.DB.ElementId
    """

    filters = view.GetFilters()
    if len(filters) != 0:
        for j in filters:
            if j not in unique_list:
                unique_list.append(j)
    return unique_list


def is_filter_applied_to_view(view, filter):
    """
    Checks if a filter is already applied to a view in Autodesk Revit.

    Args:
        view (Autodesk.Revit.DB.View): The view object to check if the filter is applied.
        filter (Autodesk.Revit.DB.Filter): The filter object to check if it is applied to the view.

    Returns:
        bool: True if the filter is applied to the view, False otherwise.
    """
    filter_is_applied = False
    filters_applied_to_view_as_ids = view.GetFilters()
    if filter.Id in filters_applied_to_view_as_ids:
        filter_is_applied = True
    return filter_is_applied


def get_filters_from_templates(doc):
    """
    Gets all filter id's used in view templates only.
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: List containing filter Id's.
    :rtype: list of Autodesk.Revit.DB.ElementId
    """

    filters_in_use = []
    # get view filters used in templates only
    # include templates which do not enforce filters but still may have some set
    template_with_filters = get_template_ids_which_can_have_filters(
        doc, VIEW_TYPE_WHICH_CAN_HAVE_FILTERS
    )
    for temp in template_with_filters:
        # get filters and check whether already in list
        filters_in_use = get_filter_ids_from_view_by_filter(temp, filters_in_use)
    return filters_in_use


def get_filter_ids_from_views_without_template(doc, filter_by_type):
    """
    Gets all filter id's from views which dont have a template applied and match a given view type.
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param filter_by_type: list of view types of which the filters are to be returned.
    :type filter_by_type: list of Autodesk.Revit.DB.ViewType
    :return: List containing filter Id's.
    :rtype: list of Autodesk.Revit.DB.ElementId
    """

    filters_in_use = []
    col = FilteredElementCollector(doc).OfClass(View)
    for v in col:
        # cant filter out templates or templates which do not control filters to be more precise
        # views The parameter:
        # BuiltInParameter.VIS_GRAPHICS_FILTERS
        # which is attached to views is of storage type None...not much use...
        if v.IsTemplate == False:
            for filter in filter_by_type:
                if v.ViewType == filter:
                    get_filter_ids_from_view_by_filter(v, filters_in_use)
                    break
    return filters_in_use


def get_all_unused_view_filters(doc):
    """
    Gets id's of all unused view filters in a model
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: List containing filter Id's.
    :rtype: list of Autodesk.Revit.DB.ElementId
    """

    un_used_view_filter_ids = []
    all_available_filters = get_all_filters(doc)
    all_filter_ids_by_template = get_filters_from_templates(doc)
    all_filter_ids_by_view = get_filter_ids_from_views_without_template(
        doc, VIEW_TYPE_WHICH_CAN_HAVE_FILTERS
    )
    # combine list of used filters into one
    all_used_view_filters = all_filter_ids_by_template + all_filter_ids_by_view
    # loop over all available filters and check for match in used filters
    for available_f in all_available_filters:
        if available_f.Id not in all_used_view_filters:
            un_used_view_filter_ids.append(available_f.Id)
    return un_used_view_filter_ids


def apply_filter_to_view(doc, view, filter):
    """
    Applies a filter to a view in Autodesk Revit.

    Args:
        doc (Autodesk.Revit.DB.Document): The document object representing the Revit model.
        view (Autodesk.Revit.DB.View): The view object to which the filter should be applied.
        filter (Autodesk.Revit.DB.Filter): The filter object that should be applied to the view.

    Returns:
        duHast.Utilities.Objects.result.Result: The result object containing the outcome of applying the filter to the view. The `result` attribute of the result object indicates whether the filter was applied successfully or not. The `messages` attribute of the result object contains any additional messages related to the application of the filter.
    """

    return_value = res.Result()

    if not is_filter_applied_to_view(view=view, filter=filter):
        # need to add filter
        return_value.append_message(
            "Filter: {} needs to be applied to view: {}".format(filter.Name, view.Name)
        )

        def action():
            action_return_value = res.Result()
            try:
                view.AddFilter(filter.Id)
                return_value.append_message(
                    "Filter: {} added to view: {} successfully.".format(
                        filter.Name, view.Name
                    )
                )
            except Exception as e:
                action_return_value.update_sep(
                    False,
                    "Failed to apply filter: {} to view: {} with exception: {}".format(
                        filter.Name,
                        view.Name,
                        e,
                    ),
                )
            return action_return_value

        transaction = Transaction(
            doc,
            "Adding filter: {} to view: {}".format(
                filter.Name,
                view.Name,
            ),
        )
        add_filter_status = in_transaction(transaction, action)
        return_value.update(add_filter_status)
    else:
        return_value.append_message(
            "Filter: {} already applied to view: {}".format(filter.Name, view.Name)
        )

    return return_value



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Revit ceilings helper functions.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

import clr
import System

# import common library modules
from duHast.Revit.Common.common import get_ids_from_element_collector
from duHast.Revit.Family.family_utils import get_all_in_place_type_ids_in_model_of_category

from duHast.Revit.Ceilings.Utility import ceilings_filter as rCeilingsFilter

# import Autodesk
from Autodesk.Revit.DB import (
    BuiltInCategory,
    Ceiling,
    ElementCategoryFilter,
    FamilyInstance,
    FilteredElementCollector,
)

# --------------------------------------------- utility functions ------------------


def get_all_ceiling_types_by_category(doc):
    """
    Gets a filtered element collector of all ceiling types in the model:

    - Compound Ceiling
    - In place families or loaded families
    - Basic Ceiling

    Filters by category.
    It will therefore not return any roof soffit types ..

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document

    :return: A filtered element collector containing ceiling types.
    :rtype: Autodesk.Revit.DB.FilteredElementCollector
    """

    collector = rCeilingsFilter._get_all_ceiling_types_by_category(doc)
    return collector


def get_ceiling_types_by_class(doc):
    """
    Gets a filtered element collector of all ceiling types in the model:

    - Roof Soffit
    - Compound Ceiling
    - Basic Ceiling

    Filters by class.
    It will therefore not return any in place family types.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document

    :return: A filtered element collector containing ceiling types.
    :rtype: Autodesk.Revit.DB.FilteredElementCollector
    """

    collector = rCeilingsFilter._get_ceiling_types_by_class(doc)
    return collector


# -------------------------------- none in place ceiling types -------------------------------------------------------


def get_all_ceiling_instances_in_model_by_category(doc):
    """
    Gets all ceiling elements placed in model. Ignores roof soffits.

    Filters by category.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document

    :return: A filtered element collector containing ceiling instances.
    :rtype: Autodesk.Revit.DB.FilteredElementCollector
    """

    return (
        FilteredElementCollector(doc)
        .OfCategory(BuiltInCategory.OST_Ceilings)
        .WhereElementIsNotElementType()
    )


def get_all_ceiling_instances_in_model_by_class(doc):
    """
    Gets all ceiling elements placed in model. Ignores in place families.

    Filters by class.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document

    :return: A filtered element collector containing ceiling instances.
    :rtype: Autodesk.Revit.DB.FilteredElementCollector
    """

    return FilteredElementCollector(doc).OfClass(Ceiling).WhereElementIsNotElementType()


def get_all_ceiling_type_ids_in_model_by_category(doc):
    """
    Gets all ceiling element type ids available in model.

    Filters by category.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document

    :return: A filtered element collector containing ceiling type ids.
    :rtype: Autodesk.Revit.DB.FilteredElementCollector
    """

    ids = []
    col_cat = get_all_ceiling_types_by_category(doc)
    ids = get_ids_from_element_collector(col_cat)
    return ids


def get_all_ceiling_type_ids_in_model_by_class(doc):
    """
    Gets all ceiling element type ids available in model.

    Filters by class.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document

    :return: A filtered element collector containing ceiling type ids.
    :rtype: Autodesk.Revit.DB.FilteredElementCollector
    """

    ids = []
    col_class = get_ceiling_types_by_class(doc)
    ids = get_ids_from_element_collector(col_class)
    return ids


# -------------------------------- In place ceiling types -------------------------------------------------------


def get_in_place_ceiling_family_instances(doc):
    """
    Gets all instances of in place families of category ceiling.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document

    :return: A filtered element collector containing in place ceiling instances.
    :rtype: Autodesk.Revit.DB.FilteredElementCollector
    """

    filter = ElementCategoryFilter(BuiltInCategory.OST_Ceilings)
    return FilteredElementCollector(doc).OfClass(FamilyInstance).WherePasses(filter)


def get_all_in_place_ceiling_type_ids_in_model(doc):
    """
    Gets all type ids off all available in place families of category ceiling.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document

    :return: List of element ids representing in place ceiling types.
    :rtype: list of Autodesk.Revit.DB.ElementId
    """

    ids = get_all_in_place_type_ids_in_model_of_category(
        doc, BuiltInCategory.OST_Ceilings
    )
    return ids



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Function to purge unused elements by delete test.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

This purge approach deletes an element from the model and checks whether that deletion changed any other elements. 
If not, means the element is unused and can be deleted.

"""

#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

# the following import statements are required to use the c# List.Any() method
import clr
import System

clr.AddReference("System.Core")
clr.ImportExtensions(System.Linq)

from duHast.Utilities.Objects.result import Result
from duHast.Utilities.Objects.timer import Timer
from duHast.Revit.Common.failure_handling import (
    get_failure_warning_report,
)

from duHast.Revit.Common.Objects.FailureHandlingConfiguration import (
    FailureHandlingConfig,
)
from duHast.Revit.Common.transaction import in_transaction_with_failure_handling

from duHast.Revit.Purge.Objects.ModifierBase import ModifierBase
from duHast.UI.Objects.ProgressBase import ProgressBase

from Autodesk.Revit.DB import (
    Element,
    Transaction,
    TransactionGroup,
    FailureProcessingResult,
)

# keeps track of number of deleted / modified elements
# 1: only the element is question got deleted...no further dependencies
# greater than 1: the element itself got deleted and as a result some other element got modified
_modified_by_delete = 0


def document_change_purge_element(
    sender, e, doc, debug, deleted_elements_modifier, modified_elements_modifier, result
):
    """
    Function to execute on DocumentChanged event and inspect the modified & deleted elements.
    An unused element will have 1 deleted element and 0 modified elements.
    """

    deleted_elements = e.GetDeletedElementIds()
    modified_elements = e.GetModifiedElementIds()
    debug_string = ""
    # check if anything got deleted or modified
    if(len(deleted_elements) == 0 and  len(modified_elements) == 0) and debug:
        debug_string += "No deleted and modified elements. Element will be ignored\n"
    else:
        try:
            if debug:
                if str(e.Operation) == "TransactionCommitted":
                    debug_string += "Deleted elements before adjustment: {}\n".format(
                        len(deleted_elements)
                    )
                    debug_string += "Modified elements before adjustment: {}\n".format(
                        len(modified_elements)
                    )

            # adjust deleted elements if necessary
            if deleted_elements_modifier != None:
                deleted_elements = deleted_elements_modifier.modify_deleted(
                    doc, deleted_elements
                )
                if debug:
                    debug_string += "Applying delete elements modifier\n"
                    debug_string += "Deleted elements after adjustment: {}\n".format(
                        len(deleted_elements)
                    )
            else:
                if debug:
                    debug_string += "No delete elements modifier provided.\n"

            # adjust deleted and modified elements if necessary:
            # this might be required where an element is presented through 2 or more elements in the revit api:
            # e.g. a line style is represented to a line style and a graphics style
            # this function will only purge elements which result in only 1 element  deleted and no other element modified
            # hence a delete modifier should check the deleted elements and if appropriate return only one element to be deleted
            # same applies to modified elements: a custom modifier should return 0 elements if appropriate in order for the element to be purged

            if modified_elements_modifier != None:
                modified_elements = modified_elements_modifier.modify_modified(
                    doc, modified_elements
                )
                if debug:
                    debug_string += "Applying modified elements modifier\n"
                    debug_string += "Modified elements after adjustment: {}\n".format(
                        len(modified_elements)
                    )
            else:
                if debug:
                    debug_string += "No modified elements modifier provided.\n"

            if debug:
                if len(modified_elements) == 0 and len(deleted_elements) == 1:
                    debug_string += "No modified elements. Element will be deleted\n"
                elif len(modified_elements) > 0:
                    debug_string += (
                        "Element will not be deleted. Event modified elements:\n"
                    )
                    for elem_id in modified_elements:
                        debug_string += "element id is of type: {}\n".format(type(elem_id))
                        elem = doc.GetElement(elem_id)
                        debug_string += "element is of type: {}\n{}\n".format(type(elem),elem)
                        try:
                            if(isinstance(elem, Element)):
                                debug_string += "{}\n".format(elem.Name)
                        except Exception as e:
                            debug_string += "An exception occurred when building debug string in modified elements: {}\n".format(e)
                            
                elif len(deleted_elements) >1:
                    debug_string += "Element will not be deleted. More than one element got deleted:\n"
                    for elem_id in deleted_elements:
                        debug_string += "Element Id: {}\n".format(
                                str(elem_id)
                            )
                        elem = doc.GetElement(elem_id)
                        debug_string += "element is of type: {}\n{}\n".format(type(elem), elem)
                        try:
                            if(isinstance(elem, Element)):
                                debug_string += "{}\n".format(elem.Name)
                        except Exception as e:
                            debug_string += "An exception occurred when building debug string in deleted elements: {}\n".format(e)
                elif len(deleted_elements) == 0 and len(modified_elements) == 0:
                    # just in case modifiers applied reduce both counts to 0...should not happen?
                    debug_string += (
                        "No deleted and modified elements. Element will be ignored\n"
                    )
                else:
                    debug_string += "Element will not be deleted. Deleted {} and modified elements: {}\n".format(
                        len(deleted_elements), len(modified_elements)
                    )

        except Exception as e:
            debug_string += "Error adjusting deleted / modified elements: {}".format(e)

    # update logs
    if debug:
        result.append_message(debug_string)
    # update global variable
    global _modified_by_delete
    _modified_by_delete = len(deleted_elements) + len(modified_elements)


def pre_process_failures(failures_accessor, process_result):
    """
    Rolls back the processing if any warning is encountered

    :param failures_accessor: The failures accessor to process
    :type failures_accessor: FailuresAccessor
    :param process_result: A Result class instance used to store any failure messages only
    :type process_result: :class:`.Result`
    :return: The result of the failure processing
    :rtype: FailureProcessingResult
    """
    try:
        # process warnings if any
        result = FailureProcessingResult.Continue
        doc = failures_accessor.GetDocument()
        app = doc.Application
        failure_registry = app.GetFailureDefinitionRegistry()
        failures = failures_accessor.GetFailureMessages()
        if failures.Any():
            process_result.append_message("")
            process_result.append_message(
                "Processing Revit document warnings / failures ({}):".format(
                    failures.Count
                )
            )
            for failure in failures:
                failure_definition = failure_registry.FindFailureDefinition(
                    failure.GetFailureDefinitionId()
                )

                # get a description of the warning which occurred
                warning_report = get_failure_warning_report(failure, failure_definition)
                process_result.append_message("\n".join(warning_report))

                # continue will roll back ( do not attempt to address the warning )
                result = FailureProcessingResult.ProceedWithRollBack
        else:
            result = FailureProcessingResult.Continue
    except Exception as e:
        process_result.update_sep(
            False, "The failure handler generated an error: {}".format(e)
        )
        result = FailureProcessingResult.Continue
    return result


def purge_unused_elements(
    doc,
    element_id_getter=None,
    deleted_elements_modifier=None,
    modified_elements_modifier=None,
    progress_callback=None,
    debug=False,
):
    """
    This script will ascertain elements not used in your model and delete them.
    It deletes the elements one by one and checks if there is modified elements
    which indicates another element was using the element. The transaction is rolled
    back if there is a modified element.

    IMPORTANT: This has flushed out some corrupt families with unresolvable geometry
    so it is advisable to run on a detached copy of your model first and rectify any
    family errors

    Credit for this script goes to Autodesk forum
    user so-chong from the below forum post
    https://forums.autodesk.com/t5/revit-api-forum/check-if-linepattern-is-in-use/td-p/7435014

    :param doc: The Revit document to purge line patterns from
    :type doc: Document
    :param DEBUG: Whether to print debug information or not. Optional. Defaults to False
    :type DEBUG: bool
    """

    return_value = Result()

    # check variables first
    if element_id_getter != None:
        if not callable(element_id_getter):
            raise TypeError("element_id_getter must be a function")
    if deleted_elements_modifier != None:
        if not isinstance(deleted_elements_modifier, ModifierBase):
            raise TypeError(
                "deleted_elements_modifier must be an instance of ModifierBase"
            )
    if modified_elements_modifier != None:
        if not isinstance(modified_elements_modifier, ModifierBase):
            raise TypeError(
                "modified_elements_modifier must be an instance of ModifierBase"
            )
    if progress_callback != None:
        if not isinstance(progress_callback, ProgressBase):
            raise TypeError("progress_callback must be an instance of ProgressBase")

    # Get the application and subscribe to the DocumentChanged event
    app = doc.Application
    # This is a lambda function that acts as the event handler.
    # It takes two parameters sender and e (representing the sender and event arguments),
    # and then it calls the document_change_purge_element function passing the sender, e, doc, and DEBUG arguments to it.
    app.DocumentChanged += lambda sender, e: document_change_purge_element(
        sender,
        e,
        doc,
        debug,
        deleted_elements_modifier,
        modified_elements_modifier,
        return_value,
    )

    # Get all element ids from getter function
    element_ids = element_id_getter(doc)

    # check if there are any elements to purge
    if len(element_ids) == 0:
        return_value.append_message("No elements found to purge")
        return return_value

    deleted_elements_data = ""
    unused_elements_count = 0
    callback_counter = 0

    # set up a timer fro each element and an overall timer
    t_by_element = Timer()
    t_overall = Timer()
    t_overall.start()

    # Loop through the line patterns and delete them one by one
    for element_id in element_ids:
        # start the timer
        t_by_element.start()
        # progress call back
        callback_counter += 1
        if progress_callback != None:
            progress_callback.update(callback_counter, len(element_ids))

        # get the actual element from the document
        element = doc.GetElement(element_id)
        # store its name for reporting later on
        element_name = Element.Name.GetValue(element)

        if debug:
            return_value.append_message("Processing element: {}".format(element_name))

        element_data = "{} (id: {})".format(element_name, element_id)
        transaction_name = "Purge element: {}".format(element_name)
        # Transaction group is what we will roll back if there is a modified element
        trans_grp = TransactionGroup(doc, transaction_name)
        trans_grp.Start()

        # set up action attempting to delete an element
        def action():
            action_return_value = Result()
            if element_id.IntegerValue < 0:
                action_return_value.append_message(
                    "Element {} {} is a built-in element and cannot be deleted".format(
                        element_id.IntegerValue, element_name
                    )
            )
            else:
                try:
                    doc.Delete(element_id)
                    action_return_value.append_message(
                        "Attempted to delete element {} :: {}".format(element_id, element_name)
                    )
                except Exception as e:
                    action_return_value.update_sep(False,
                        "Element {} could not be deleted: {}".format(element_name, e)
                    )
                    # note if an exception is thrown at delete step
                    # the modified_by_delete will remain 0 and the transaction will be rolled back
                    # document change event does not get triggered
            return action_return_value
        
        # define failure handling for the transaction ( roll back on any warnings or errors )
        failure_handling_settings = FailureHandlingConfig(
            roll_back_on_warning=True,
            print_warnings=False,
            roll_back_on_error=True,
            print_errors=False,
        )

        # set up the transaction
        trans = Transaction(doc, transaction_name)

        # execute the transaction with failure handling
        result_transaction = in_transaction_with_failure_handling(
            transaction=trans,
            action=action,
            failure_config=failure_handling_settings,
            failure_processing_func=pre_process_failures,
        )

        # update the return value with the result of the transaction
        return_value.update(result_transaction)

        global _modified_by_delete

        if _modified_by_delete == 1:
            unused_elements_count += 1
            deleted_elements_data += element_data + "\n"
            trans_grp.Assimilate()

        else:
            trans_grp.RollBack()

        # Reset the modified_by_delete variable here incase it is
        # not reset in the DocumentChanged event
        _modified_by_delete = 0

        return_value.append_message(
            "Element {} took: {}".format(element_id, t_by_element.stop())
        )

    deleted_elements_data += "\n"
    return_value.append_message(
        "\nDeleted {} unused elements:\n\n{}".format(
            unused_elements_count, deleted_elements_data
        )
    )

    return_value.append_message("Elements before purge: {}\n".format(len(element_ids)))
    # Get all line patterns in the model again for a post-task count
    element_ids = element_id_getter(doc)
    return_value.append_message("Elements after purge: {}".format(len(element_ids)))

    # Unsubscribe from the DocumentChanged event
    app.DocumentChanged -= lambda sender, e: document_change_purge_element(
        sender, e, doc, debug
    )

    return_value.append_message("Purge took: {}".format(t_overall.stop()))
    return return_value



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
This module contains a purge unused function using standard revit api functions.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Some functionality provided here exceeds the Revit purge unused or e-transmit purge unused command:

- view types
- view templates
- view filters

Others definitely lack:

- Materials
- Appearance assets
- loadable Families
- some MEP systems

Future: just provide improvements over e-transmit purge unused in this code section.

"""

# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed. 
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits; 
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

import clr
import System

from duHast.Utilities import files_io as fileIO
from duHast.Utilities.Objects import result as res

from duHast.Revit.Annotation import spot_dimensions as rAnnoSpot
from duHast.Revit.Annotation import generic_annotation as rGAnno

from duHast.Revit.Family import family_utils as rFamUPurge
from duHast.Revit.Family import purge_unused_family_types as rFamPurge

from duHast.Revit.Common import groups as rGrp
from duHast.Revit.Common import delete as rDel

from duHast.Revit.Views import views as rView
from duHast.Revit.Views import referencing as rViewRef
from duHast.Revit.Views import filters as rViewFilter
from duHast.Revit.Views import views_purge_unused as rViewPurge
from duHast.Revit.Views import templates as rViewTemp

from duHast.Revit.Walls import walls as rWall
from duHast.Revit.Walls import purge_unused_wall_types as rWallPurge
from duHast.Revit.Walls import curtain_wall_elements as rCurtainWallElem
from duHast.Revit.Walls import curtain_walls as rCurtainWall
from duHast.Revit.Walls import stacked_walls as rStackedWall
from duHast.Revit.Walls import (
    purge_unused_curtain_wall_element_types as rCurtainWallElemPurge,
)

from duHast.Revit.Annotation import purge_unused_annotation_types as rAnnoPurge
from duHast.Revit.Annotation import arrow_heads as rArrow
from duHast.Revit.Annotation import dimensions as rDim
from duHast.Revit.Annotation import multi_ref_annotation as rMultiRefAnno
from duHast.Revit.Annotation import text as rText

from duHast.Revit.Stairs import purge_unused_stair_types as rStairPurge
from duHast.Revit.Stairs import stairs as rStair
from duHast.Revit.Stairs import runs as rStairRun
from duHast.Revit.Stairs import landings as rStairLanding
from duHast.Revit.Stairs import path as rStairPath
from duHast.Revit.Stairs import stringers_carriages as rStairStringerAndCarriage
from duHast.Revit.Stairs import cut_marks as rStairCutMark

from duHast.Revit.BuildingPads import building_pads as rBuildP
from duHast.Revit.BuildingPads import (
    purge_unused_building_pad_types as rBuildingPadPurge,
)

from duHast.Revit.Ceilings import ceilings as rCeil
from duHast.Revit.Ceilings import purge_unused_ceiling_types as rCeilingPurge

from duHast.Revit.DetailItems import detail_items as rDet
from duHast.Revit.DetailItems import purge_unused_detail_item_types as rDetailItemPurge

from duHast.Revit.Floors import floors as rFlo
from duHast.Revit.Floors import purge_unused_floor_types as rFloorPurge

from duHast.Revit.Grids import grids as rGrid
from duHast.Revit.Grids import purge_unused_grid_types as rGridPurge

from duHast.Revit.Levels import levels as rLev
from duHast.Revit.Levels import purge_unused_level_types as rLevelPurge

from duHast.Revit.Links import image_links as rImageLink
from duHast.Revit.Links import purge_unused_image_link_types as rLinkPurge

from duHast.Revit.Railings import railings as rRail
from duHast.Revit.Railings import balusters as rBal
from duHast.Revit.Railings import purge_unused_railing_and_baluster_types as rRailPurge

from duHast.Revit.MEP_Systems import purge_unused_mep_symbols as rMEPSymbolPurge
from duHast.Revit.MEP_Systems import purge_unused_mep_types as rMEPTypePurge
from duHast.Revit.MEP_Systems import cable_trays as rCableTray
from duHast.Revit.MEP_Systems import conduits as rConduit
from duHast.Revit.MEP_Systems import ducts as rDuct
from duHast.Revit.MEP_Systems import pipes as rPipe

from duHast.Revit.Ramps import ramps as rRam
from duHast.Revit.Ramps import purge_unused_ramp_types as rRampPurge

from duHast.Revit.Roofs import roofs as rRoof
from duHast.Revit.Roofs import purge_unused_roof_types as rRoofPurge

from duHast.Utilities.Objects.timer import Timer
from duHast.Revit.Purge import purge_action as pA

import Autodesk.Revit.DB as rdb
from System.Collections.Generic import List


# --------------------------------------------- Purge - utility ---------------------------------------------


def purge_unplaced_elements(
    doc,
    get_unused_element_ids,
    transaction_name,
    unused_element_name_header,
    is_debug=False,
):
    """
    
    Purges all unplaced elements provided through a passed in element id getter method from a model.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param get_unused_element_ids: A function accepting the current document as the argument and returning element ids which can be purged.
    :type get_unused_element_ids: func (doc) returning list of Autodesk.Revit.DB.ElementId
    :param transaction_name: A human readable description of the transaction containing the purge action.
    :type transaction_name: str
    :param unused_element_name_header: The text to be displayed at the start of the list containing the deleted element names.
    :type unused_element_name_header: str
    :param is_debug: True: will return detailed report and attempt to try to delete elements one by one if an exception occurs, defaults to False\
        Will attempt to delete all elements at once.
    :type is_debug: bool, optional

    :return: 
        Result class instance.
        
        - .status True if all purge actions completed successfully. Otherwise False.
        - .message will be listing each purge action and its status
    
    :rtype: :class:`.Result`
    """

    resultValue = res.Result()
    try:
        unusedElementIds = get_unused_element_ids(doc)
        unusedElementNames = []
        if is_debug:
            unusedElementNames.append(unused_element_name_header)
            for unusedId in unusedElementIds:
                unusedElementNames.append(
                    SPACER
                    + "ID:\t"
                    + str(unusedId)
                    + " Name:\t"
                    + rdb.Element.Name.GetValue(doc.GetElement(unusedId))
                )
        else:
            unusedElementNames.append(
                unused_element_name_header
                + ": "
                + str(len(unusedElementIds))
                + " Element(s) purged."
            )
        purgeResult = rDel.delete_by_element_ids(
            doc, unusedElementIds, transaction_name, "\n".join(unusedElementNames)
        )
        # check if an exception occurred and in debug mode, purge elements one by one
        if is_debug and purgeResult.status == False:
            # pass
            print("second debug run")
            purgeResult = rDel.delete_by_element_ids_one_by_one(
                doc, unusedElementIds, transaction_name, "\n".join(unusedElementNames)
            )
        resultValue.update(purgeResult)
    except Exception as e:
        resultValue.update_sep(
            False,
            "Terminated purge unused "
            + unused_element_name_header
            + " with exception: "
            + str(e),
        )
    return resultValue


# --------------------------------------------- Main ---------------------------------------------

#: list containing purge action names and the purge action method
PURGE_ACTIONS = []
PURGE_ACTIONS.append(
    pA.PurgeAction(
        "Purge Unused Model Group(s)",
        rGrp.get_unplaced_model_group_ids,
        "Model Group(s)",
        "Model Group(s)",
        rGrp.get_model_group_ids,
    )
)
PURGE_ACTIONS.append(
    pA.PurgeAction(
        "Purge Unused Detail Group(s)",
        rGrp.get_unplaced_detail_group_ids,
        "Detail Group(s)",
        "Detail Group(s)",
        rGrp.get_detail_group_ids,
    )
)
PURGE_ACTIONS.append(
    pA.PurgeAction(
        "Purge Unused Nested Detail Group(s)",
        rGrp.get_unplaced_nested_detail_group_ids,
        "Nested Detail Group(s)",
        "Nested Detail Group(s)",
        rGrp.get_nested_detail_group_ids,
    )
)
PURGE_ACTIONS.append(
    pA.PurgeAction(
        "Purge Unused View Family Types",
        rViewPurge.get_unused_view_type_ids,
        "View Family Type(s)",
        "View Family Type(s)",
        rView.get_view_type_ids,
    )
)
PURGE_ACTIONS.append(
    pA.PurgeAction(
        "Purge Unused View Templates",
        rViewTemp.get_all_unused_view_template_ids,
        "View Family Templates(s)",
        "View Family Templates(s)",
        rViewTemp.get_view_templates_ids,
    )
)
PURGE_ACTIONS.append(
    pA.PurgeAction(
        "Purge Unused View Filters",
        rViewFilter.get_all_unused_view_filters,
        "View Filter(s)",
        "View Filter(s)",
        rViewFilter.get_all_filter_ids,
    )
)
PURGE_ACTIONS.append(
    pA.PurgeAction(
        "Purge Unused Image Links",
        rLinkPurge.get_all_unused_image_type_ids_in_model,
        "Images(s)",
        "Images(s)",
        rImageLink.get_images_type_ids_in_model,
    )
)
PURGE_ACTIONS.append(
    pA.PurgeAction(
        "Purge Unused Stacked Wall Types",
        rWallPurge.get_unused_stacked_wall_type_ids_to_purge,
        "Stacked Wall Type(s)",
        "Stacked Wall Type(s)",
        rStackedWall.get_all_stacked_wall_type_ids,
    )
)
PURGE_ACTIONS.append(
    pA.PurgeAction(
        "Purge Unused InPlace Wall Types",
        rWallPurge.get_unused_in_place_wall_ids_for_purge,
        "InPlace Wall Type(s)",
        "InPlace Wall Type(s)",
        rWall.get_all_in_place_wall_type_ids,
    )
)
PURGE_ACTIONS.append(
    pA.PurgeAction(
        "Purge Unused Curtain Wall Types",
        rWallPurge.get_unused_curtain_wall_type_ids_to_purge,
        "Curtain Wall Type(s)",
        "Curtain Wall Type(s)",
        rCurtainWall.get_all_curtain_wall_type_ids,
    )
)
PURGE_ACTIONS.append(
    pA.PurgeAction(
        "Purge Unused Basic Types",
        rWallPurge.get_unused_basic_wall_type_ids_to_purge,
        "Basic Wall Type(s)",
        "Basic Wall Type(s)",
        rWall.get_all_basic_wall_type_ids,
    )
)
PURGE_ACTIONS.append(
    pA.PurgeAction(
        "Purge Unused Curtain Wall Element Types",
        rCurtainWallElemPurge.get_unused_non_symbol_curtain_wall_element_type_ids_to_purge,
        "Curtain Wall Element Type(s)",
        "Curtain Wall Element Type(s)",
        rCurtainWallElem.get_all_curtain_wall_element_type_ids_by_category_excl_symbols,
    )
)
PURGE_ACTIONS.append(
    pA.PurgeAction(
        "Purge Unused Loadable Curtain Wall Symbol (Types)",
        rCurtainWallElemPurge.get_unused_curtain_wall_symbol_ids_for_purge,
        "Curtain Wall Loadable Symbols (Type(s))",
        "Curtain Wall Loadable Symbols (Type(s))",
        rCurtainWallElem.get_all_curtain_wall_non_shared_symbol_ids_by_category,
    )
)
PURGE_ACTIONS.append(
    pA.PurgeAction(
        "Purge Unused Ceiling Types",
        rCeilingPurge.get_unused_non_in_place_ceiling_type_ids_to_purge,
        "Ceiling Type(s)",
        "Ceiling Type(s)",
        rCeil.get_all_ceiling_type_ids_in_model_by_class,
    )
)  # used by class filter to avoid in place families listed
PURGE_ACTIONS.append(
    pA.PurgeAction(
        "Purge Unused InPlace Ceiling Types",
        rCeilingPurge.get_unused_in_place_ceiling_ids_for_purge,
        "InPlace Ceiling Type(s)",
        "InPlace Ceiling Type(s)",
        rCeil.get_all_in_place_ceiling_type_ids_in_model,
    )
)
PURGE_ACTIONS.append(
    pA.PurgeAction(
        "Purge Unused Floor Types",
        rFloorPurge.get_unused_non_in_place_floor_type_ids_to_purge,
        "Floor Type(s)",
        "Floor Type(s)",
        rFlo.get_all_floor_type_ids_in_model_by_class,
    )
)  # TODO check why this is using by class...
PURGE_ACTIONS.append(
    pA.PurgeAction(
        "Purge Unused InPlace Floor Types",
        rFloorPurge.get_unused_in_place_floor_ids_for_purge,
        "InPlace Floor Type(s)",
        "InPlace Floor Type(s)",
        rFlo.get_all_in_place_floor_type_ids_in_model,
    )
)
PURGE_ACTIONS.append(
    pA.PurgeAction(
        "Purge Unused Roof Types",
        rRoofPurge.get_unused_non_in_place_roof_type_ids_to_purge,
        "Roof Type(s)",
        "Roof Type(s)",
        rRoof.get_all_roof_type_ids_by_class,
    )
)  # TODO check why by class
PURGE_ACTIONS.append(
    pA.PurgeAction(
        "Purge Unused InPlace Roof Types",
        rRoofPurge.get_unused_in_place_roof_type_ids_for_purge,
        "InPlace Roof Type(s)",
        "InPlace Roof Type(s)",
        rRoof.get_all_in_place_roof_type_ids,
    )
)
PURGE_ACTIONS.append(
    pA.PurgeAction(
        "Purge Unused Stair Types",
        rStairPurge.get_unused_non_in_place_stair_type_ids_to_purge,
        "Stair Type(s)",
        "Stair Type(s)",
        rStair.get_all_stair_type_ids_by_class,
    )
)  # TODO check why by class
PURGE_ACTIONS.append(
    pA.PurgeAction(
        "Purge Unused Path Types",
        rStairPurge.get_unused_stair_path_type_ids_to_purge,
        "Stair Path Type(s)",
        "Stair Path Type(s)",
        rStairPath.get_stair_path_types_ids_by_class,
    )
)
PURGE_ACTIONS.append(
    pA.PurgeAction(
        "Purge Unused Landing Types",
        rStairPurge.get_unused_stair_landing_type_ids_to_purge,
        "Stair Landing Type(s)",
        "Stair Landing Type(s)",
        rStairLanding.get_stair_landing_types_ids_by_class,
    )
)
PURGE_ACTIONS.append(
    pA.PurgeAction(
        "Purge Unused Run Types",
        rStairPurge.get_unused_stair_run_type_ids_to_purge,
        "Stair Run Type(s)",
        "Stair Run Type(s)",
        rStairRun.get_stair_run_types_ids_by_class,
    )
)
PURGE_ACTIONS.append(
    pA.PurgeAction(
        "Purge Unused Stringers and Carriage Types",
        rStairPurge.get_unused_stair_stringers_carriage_type_ids_to_purge,
        "Stair Stringers and Carriage Type(s)",
        "Stair Stringers and Carriage Type(s)",
        rStairStringerAndCarriage.get_all_stair_stringers_carriage_type_ids_by_category,
    )
)
PURGE_ACTIONS.append(
    pA.PurgeAction(
        "Purge Unused InPlace Stair Types",
        rStairPurge.get_unused_in_place_stair_type_ids_for_purge,
        "InPlace Stair Type(s)",
        "InPlace Stair Type(s)",
        rStair.get_all_in_place_stair_type_ids,
    )
)
PURGE_ACTIONS.append(
    pA.PurgeAction(
        "Purge Unused Ramp Types",
        rRampPurge.get_unused_non_in_place_ramp_type_ids_to_purge,
        "Ramp Type(s)",
        "Ramp Type(s)",
        rRam.get_all_ramp_types_ids_by_category,
    )
)
PURGE_ACTIONS.append(
    pA.PurgeAction(
        "Purge Unused Stair Cut Mark Types",
        rStairPurge.get_unused_stair_cut_mark_type_ids_to_purge,
        "Stair Cut Mark Type(s)",
        "Stair Cut Mark Type(s)",
        rStairCutMark.get_stair_cut_mark_types_ids_by_class,
    )
)
PURGE_ACTIONS.append(
    pA.PurgeAction(
        "Purge Unused Building Pad Types",
        rBuildingPadPurge.get_unused_non_in_place_building_pad_type_ids_to_purge,
        "Building Pad Type(s)",
        "Building Pad Type(s)",
        rBuildP.get_all_building_pad_type_ids_in_model_by_class,
    )
)
PURGE_ACTIONS.append(
    pA.PurgeAction(
        "Purge Unused Railing Types",
        rRailPurge.get_unused_non_in_place_railing_type_ids_to_purge,
        "Railing Type(s)",
        "Railing Type(s)",
        rRail.get_all_railing_type_ids_by_class_and_category,
    )
)
PURGE_ACTIONS.append(
    pA.PurgeAction(
        "Purge Unused InPlace Railing Types",
        rRailPurge.get_unused_in_place_railing_ids_for_purge,
        "In Place Railing Type(s)",
        "In Place Railing Type(s)",
        rRail.get_in_place_railing_type_ids_in_model,
    )
)
PURGE_ACTIONS.append(
    pA.PurgeAction(
        "Purge Unused Baluster Types",
        rRailPurge.get_unused_baluster_type_ids_for_purge,
        "Baluster Type(s)",
        "Baluster Type(s)",
        rBal.get_all_baluster_symbols_ids,
    )
)
PURGE_ACTIONS.append(
    pA.PurgeAction(
        "Purge Unused Cable Tray Types",
        rMEPTypePurge.get_unused_cable_tray_type_ids_to_purge,
        "Cable Tray Type(s)",
        "Cable Tray Type(s)",
        rCableTray.get_all_cable_tray_type_ids_in_model_by_category,
    )
)
PURGE_ACTIONS.append(
    pA.PurgeAction(
        "Purge Unused Conduit Types",
        rMEPTypePurge.get_unused_conduit_type_ids_to_purge,
        "Conduit Type(s)",
        "Conduit Type(s)",
        rConduit.get_all_conduit_type_ids_in_model_by_category,
    )
)
PURGE_ACTIONS.append(
    pA.PurgeAction(
        "Purge Unused Duct Types",
        rMEPTypePurge.get_unused_duct_type_ids_to_purge,
        "Duct Type(s)",
        "Duct Type(s)",
        rDuct.get_all_duct_type_ids_in_model_by_category,
    )
)
PURGE_ACTIONS.append(
    pA.PurgeAction(
        "Purge Unused Pipe Types",
        rMEPTypePurge.get_unused_pipe_type_ids_to_purge,
        "Pipe Type(s)",
        "Pipe Type(s)",
        rPipe.get_all_pipe_type_ids_in_model_by_category,
    )
)
PURGE_ACTIONS.append(
    pA.PurgeAction(
        "Purge Unused Cable Tray Symbols and Families",
        rMEPSymbolPurge.get_unused_cable_tray_symbol_ids_for_purge,
        "Cable Tray Symbols and Family(s)",
        "Cable Tray Symbols and Family(s)",
        rCableTray.get_symbol_ids_for_cable_tray_types_in_model,
    )
)
PURGE_ACTIONS.append(
    pA.PurgeAction(
        "Purge Unused Conduit Symbols and Families",
        rMEPSymbolPurge.get_unused_conduit_symbol_ids_for_purge,
        "Conduit Symbols and Family(s)",
        "Conduit Symbols and Family(s)",
        rConduit.get_symbol_ids_for_conduit_types_in_model,
    )
)
PURGE_ACTIONS.append(
    pA.PurgeAction(
        "Purge Unused Duct Symbols and Families",
        rMEPSymbolPurge.get_unused_duct_and_flex_duct_symbol_ids_for_purge,
        "Duct Symbols and Family(s)",
        "Duct Symbols and Family(s)",
        rDuct.get_symbol_ids_for_duct_types_in_model,
    )
)
PURGE_ACTIONS.append(
    pA.PurgeAction(
        "Purge Unused Pipe Symbols and Families",
        rMEPSymbolPurge.get_unused_pipe_symbol_ids_for_purge,
        "Pipe Symbols and Family(s)",
        "Pipe Symbols and Family(s)",
        rPipe.get_symbol_ids_for_pipe_types_in_model,
    )
)
PURGE_ACTIONS.append(
    pA.PurgeAction(
        "Purge Unused Level Types",
        rLevelPurge.get_unused_level_types_for_purge,
        "Level Type(s)",
        "Level Type(s)",
        rLev.get_all_level_type_ids_by_category,
    )
)
PURGE_ACTIONS.append(
    pA.PurgeAction(
        "Purge Unused Level Head Types",
        rLevelPurge.get_unused_level_head_families_for_purge,
        "Level Head family Type(s)",
        "Level Head family Type(s)",
        rLev.get_all_level_head_family_type_ids,
    )
)
PURGE_ACTIONS.append(
    pA.PurgeAction(
        "Purge Unused Grid Types",
        rGridPurge.get_unused_grid_types_for_purge,
        "Grid Type(s)",
        "Grid Type(s)",
        rGrid.get_all_grid_type_ids_by_category,
    )
)
PURGE_ACTIONS.append(
    pA.PurgeAction(
        "Purge Unused Grid Head Types",
        rGridPurge.get_unused_grid_head_families_for_purge,
        "Grid Head family Type(s)",
        "Grid Head family Type(s)",
        rGrid.get_all_grid_head_family_type_ids,
    )
)
PURGE_ACTIONS.append(
    pA.PurgeAction(
        "Purge Unused View Reference Types",
        rViewPurge.get_unused_view_reference_type_ids_for_purge,
        "View Ref Type(s)",
        "View Ref Type(s)",
        rViewRef.get_all_view_reference_type_id_data_as_list,
    )
)
PURGE_ACTIONS.append(
    pA.PurgeAction(
        "Purge Unused View Continuation Types",
        rViewPurge.get_unused_continuation_marker_type_ids_for_purge,
        "View Continuation Type(s)",
        "View Continuation Type(s)",
        rViewRef.get_all_view_continuation_type_ids,
    )
)
PURGE_ACTIONS.append(
    pA.PurgeAction(
        "Purge Unused View Reference Families",
        rViewPurge.get_unused_view_ref_and_continuation_marker_families_for_purge,
        "View Ref and Continuation Marker families(s)",
        "View Ref and Continuation Marker families(s)",
        rViewRef.get_all_view_reference_symbol_ids,
    )
)
PURGE_ACTIONS.append(
    pA.PurgeAction(
        "Purge Unused Repeating Details",
        rDetailItemPurge.get_unused_repeating_detail_type_ids_for_purge,
        "Repeating Detail Type(s)",
        "Repeating Detail Type(s)",
        rDet.get_all_repeating_detail_type_ids_available,
    )
)
PURGE_ACTIONS.append(
    pA.PurgeAction(
        "Purge Unused Filled Regions",
        rDetailItemPurge.get_unused_filled_region_type_ids_for_purge,
        "Filled Region Type(s)",
        "Filled Region Type(s)",
        rDet.get_all_filled_region_type_ids_available,
    )
)
PURGE_ACTIONS.append(
    pA.PurgeAction(
        "Purge Unused Details Symbols",
        rDetailItemPurge.get_all_unused_detail_symbol_ids_for_purge,
        "Detail Symbol(s)",
        "Detail Symbol(s)",
        rDet.get_all_detail_symbol_ids_available,
    )
)
PURGE_ACTIONS.append(
    pA.PurgeAction(
        "Purge Unused MultiRef Dimension Types",
        rAnnoPurge.get_all_unused_multi_ref_dim_type_ids_in_model,
        "MultiRef Dimension Type(s)",
        "MultiRef Dimension Type(s)",
        rMultiRefAnno.get_all_multi_ref_annotation_type_ids,
    )
)
PURGE_ACTIONS.append(
    pA.PurgeAction(
        "Purge Unused Dimension Types",
        rAnnoPurge.get_all_unused_dim_type_ids_in_model,
        "Dimension Type(s)",
        "Dimension Type(s)",
        rDim.get_dim_type_ids,
    )
)
PURGE_ACTIONS.append(
    pA.PurgeAction(
        "Purge Unused Text Types",
        rAnnoPurge.get_all_unused_text_type_ids_in_model,
        "Text Type(s)",
        "Text Type(s)",
        rText.get_all_text_type_ids,
    )
)
PURGE_ACTIONS.append(
    pA.PurgeAction(
        "Purge Unused Arrow Heads",
        rAnnoPurge.get_all_unused_arrow_type_ids_in_model,
        "Arrow Head Type(s)",
        "Arrow Head Type(s)",
        rArrow.get_arrow_type_ids_in_model,
    )
)
PURGE_ACTIONS.append(
    pA.PurgeAction(
        "Purge Unused Generic Annotation",
        rAnnoPurge.get_unused_generic_annotation_ids_for_purge,
        "Generic Anno Type(s)",
        "Generic Anno Type(s)",
        rGAnno.get_all_generic_annotation_type_ids_by_category,
    )
)
PURGE_ACTIONS.append(
    pA.PurgeAction(
        "Purge Unused SpotElevation Symbols",
        rAnnoPurge.get_unused_symbol_ids_from_spot_types_to_purge,
        "Spot Elevation Symbol(s)",
        "Spot Elevation Symbol(s)",
        rAnnoSpot.get_all_spot_elevation_symbol_ids_in_model,
    )
)
PURGE_ACTIONS.append(
    pA.PurgeAction(
        "Purge Unused Loadable Family Types",
        rFamPurge.get_unused_non_shared_family_symbols_and_type_ids_to_purge,
        "Loadable Non Shared Family Type(s)",
        "Loadable Non Shared Family Type(s)",
        rFamUPurge.get_all_non_shared_family_symbol_ids,
    )
)  # TODO check its not deleting to much


#: indentation for names of items purged
SPACER = "..."


def purge_unused(doc, revit_file_path, is_debug):
    """
    Calls all available purge actions defined in global list.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param revit_file_path: Fully qualified file path of current model document. (Not used)
    :type revit_file_path: str
    :param is_debug: True: will return detailed report and attempt to try to delete elements one by one if an exception occurs. False\
        Will attempt to delete all elements at once, less detailed purge report.
    :type is_debug: bool

    :return: 
        Result class instance.
        
        - .status True if all purge actions completed successfully. Otherwise False.
        - .message will be listing each purge action and its status
    
    :rtype: :class:`.Result`
    """

    # the current file name
    revitFileName = fileIO.get_file_name_without_ext(revit_file_path)
    result_value = res.Result()
    # set up a timer objects
    timer_task = Timer()
    timer_overall = Timer()
    timer_overall.start()
    for pA in PURGE_ACTIONS:
        try:
            timer_task.start()
            purge_flag = purge_unplaced_elements(
                doc,
                pA.purge_ids_getter,
                pA.purge_transaction_name,
                pA.purge_report_header,
                is_debug,
            )
            purge_flag.append_message("{}{}".format(SPACER, timer_task.stop()))
            result_value.update(purge_flag)
        except Exception as e:
            timer_task.stop()
            result_value.update_sep(
                False, "Terminated purge unused actions with exception: {}".format(e)
            )
    result_value.append_message("purge duration: {}".format(timer_overall.stop()))
    return result_value



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Interface for family data storage / processing class.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""

#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed. 
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits; 
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

from duHast.Utilities.Objects import base

# common data dictionary keys
ROOT = "root"
ROOT_CATEGORY = "rootCategory"
FAMILY_NAME = "familyName"
FAMILY_FILE_PATH = "familyFilePath"
USAGE_COUNTER = "usageCounter"
USED_BY = "usedBy"


class IFamilyData(base.Base):
    def __init__(self, root_path, root_category_path=None, data_type=None, **kwargs):

        # forwards all unused arguments
        # ini super class to allow multi inheritance in children!
        super(IFamilyData, self).__init__(**kwargs)

        self.data = []

        if data_type != None:
            self.data_type = data_type
        else:
            self.data_type = "not declared"

        if root_path != None:
            self.root_path = root_path
        else:
            self.root_path = "-"

        if root_category_path != None:
            self.root_category_path = root_category_path
        else:
            self.root_category_path = "-"

    def process(self, doc):
        pass

    def get_data(self):
        pass

    def update_data(
        self,
        identify_by_this_property_name,
        identify_by_this_property_value,
        update_dic,
    ):
        match = False
        match_update = True
        for d in self.data:
            # print(identifyByThisPropertyName, d)
            if identify_by_this_property_name in d:
                # print('identify by property found')
                if d[identify_by_this_property_name] == identify_by_this_property_value:
                    # print ('dic', updateDic)
                    for update_prop in update_dic:
                        if update_prop in d:
                            old_value = d[update_prop]
                            d[update_prop] = update_dic[update_prop]
                            # print ('updated:', update_prop, ' from value ', old_value, ' to value ', d[update_prop])
                            match_update = match_update and True

        if match_update:
            return match_update
        else:
            return match

    def add_data(self):
        pass

    def _strip_file_extension(self, fam_name):
        """
        Strips the file extension '.rfa. , if exists, of the family  name.

        :param famName: The family name.
        :type famName: str
        :return: The truncated family name.
        :rtype: str
        """

        if fam_name.lower().endswith(".rfa"):
            fam_name = fam_name[:-4]
        return fam_name



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
This module contains a Revit rooms geometry extraction functions. 
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed. 
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits; 
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

import Autodesk.Revit.DB as rdb

from duHast.Revit.Rooms.rooms import get_all_rooms
from duHast.Data.Objects.Properties.Geometry import geometry_polygon as dGeometryPoly


def get_room_boundary_loops(revit_room,
                            spatial_boundary_option=rdb.SpatialElementBoundaryOptions(),
                            boundary_location=rdb.SpatialElementBoundaryLocation.Center):
    """
    Returns all boundary loops for a rooms. Default value set to the center
    boundary location.

    :param revit_room: The room.
    :type revit_room: Autodesk.Revit.DB.Architecture.Room
    :return: List of boundary loops defining the room.
    :rtype: List of lists of Autodesk.Revit.DB.BoundarySegment
    """

    all_boundary_loops = []
    # set up spatial boundary option
    spatial_boundary_option.StoreFreeBoundaryFaces = True
    spatial_boundary_option.SpatialElementBoundaryLocation = boundary_location
    # get loops
    loops = revit_room.GetBoundarySegments(spatial_boundary_option)
    all_boundary_loops.append(loops)
    return all_boundary_loops


def get_points_from_room_boundaries(boundary_loops):
    """
    Returns a list of lists of points representing the room boundary loops.

    - List of Lists because a room can be made up of multiple loops (holes in rooms!)
    - First nested list represents the outer boundary of a room
    - All loops are implicitly closed ( last point is not the first point again!)
    
    :param boundary_loops: List of boundary loops defining the room.
    :type boundary_loops: List of lists of Autodesk.Revit.DB.BoundarySegment
    :return: A data geometry instance containing the points defining the boundary loop.
    :rtype: :class:`.DataPolygon`
    """

    loop_counter = 0
    has_inner_loops = False
    data_geo_polygon = dGeometryPoly.DataPolygon()
    for boundary_loop in boundary_loops:
        for room_loop in boundary_loop:
            p = None  # segment start point
            loop_points = []
            for segment in room_loop:
                p = segment.GetCurve().GetEndPoint(0)
                loop_points.append(p)
            if loop_counter == 0:
                data_geo_polygon.outer_loop = loop_points
            else:
                data_geo_polygon.inner_loops.append(loop_points)
                has_inner_loops = True
            loop_counter += 1
    if not has_inner_loops:
        data_geo_polygon.inner_loops = []
    return data_geo_polygon


def get_2d_points_from_revit_room(revit_room):
    """
    Returns a list of dataGeometry object containing points representing the flattened(2D geometry) of a room in the model.
    List should only have one entry.
    :param revit_room: The room.
    :type revit_room: Autodesk.Revit.DB.Architecture.Room
    :return: A list of data geometry instance containing the points defining the boundary loop.
    :rtype: list of  :class:`.DataGeometry`
    """

    all_room_points = []
    boundary_loops = get_room_boundary_loops(revit_room)
    if len(boundary_loops) > 0:
        room_points = get_points_from_room_boundaries(boundary_loops)
        all_room_points.append(room_points)
    return all_room_points


def get_2d_points_from_all_revit_rooms(doc):
    """
    Returns a list of dataGeometry object containing points representing the flattened(2D geometry) of all the rooms in the model.
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: A list of data geometry instances containing the points defining the boundary loop per room.
    :rtype: list of  :class:`.DataGeometry`
    """

    all_room_point_groups = []
    rooms = get_all_rooms(doc)
    for room in rooms:
        room_points = get_2d_points_from_revit_room(room)
        if len(room_points) > 0:
            all_room_point_groups.append(room_points)
    return all_room_point_groups
